import platform as platform_module
import subprocess
import sys
import typing
from contextlib import contextmanager
from pathlib import Path, PurePosixPath
from unittest import mock

import pytest

from cibuildwheel import platforms
from cibuildwheel.__main__ import main
from cibuildwheel.oci_container import OCIPlatform
from cibuildwheel.util import file

DEFAULT_IDS = {"cp38", "cp39", "cp310", "cp311", "cp312", "cp313", "cp314", "cp314t"}
ALL_IDS = DEFAULT_IDS | {"cp313t", "pp38", "pp39", "pp310", "pp311", "gp311_242", "gp312_250"}


@pytest.fixture
def mock_build_container(monkeypatch: pytest.MonkeyPatch) -> None:
    def fail_on_call(*args: object, **kwargs: object) -> None:
        msg = "This should never be called"
        raise RuntimeError(msg)

    def ignore_call(*args: object, **kwargs: object) -> None:
        pass

    @contextmanager
    def nullcontext(enter_result: object = None) -> typing.Generator[object, None, None]:
        yield enter_result

    def ignore_context_call(*args: object, **kwargs: object) -> typing.ContextManager[object]:
        return nullcontext(kwargs)

    monkeypatch.setenv("CIBW_PLATFORM", "linux")
    monkeypatch.setattr(platform_module, "machine", lambda: "x86_64")

    monkeypatch.setattr(subprocess, "Popen", fail_on_call)
    monkeypatch.setattr(subprocess, "run", ignore_call)
    monkeypatch.setattr(file, "download", fail_on_call)
    monkeypatch.setattr("cibuildwheel.platforms.linux.OCIContainer", ignore_context_call)

    monkeypatch.setattr(
        "cibuildwheel.platforms.linux.build_in_container",
        mock.Mock(spec=platforms.linux.build_in_container),
    )
    monkeypatch.setattr("cibuildwheel.logger.Logger.print_summary", ignore_context_call)


@pytest.mark.usefixtures("mock_build_container", "fake_package_dir")
def test_build_default_launches(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(sys, "argv", [*sys.argv, "--platform=linux"])
    monkeypatch.setenv("CIBW_ARCHS", "auto64 auto32")
    monkeypatch.delenv("CIBW_ENABLE", raising=False)

    main()

    build_in_container = typing.cast("mock.Mock", platforms.linux.build_in_container)

    assert build_in_container.call_count == 4

    # In Python 3.8+, this can be simplified to [0].kwargs
    kwargs = build_in_container.call_args_list[0][1]
    assert "quay.io/pypa/manylinux_2_28_x86_64" in kwargs["container"]["image"]
    assert kwargs["container"]["cwd"] == PurePosixPath("/project")
    assert kwargs["container"]["oci_platform"] == OCIPlatform.AMD64

    identifiers = {x.identifier for x in kwargs["platform_configs"]}
    assert identifiers == {f"{x}-manylinux_x86_64" for x in DEFAULT_IDS}

    kwargs = build_in_container.call_args_list[1][1]
    assert "quay.io/pypa/manylinux_2_28_i686" in kwargs["container"]["image"]
    assert kwargs["container"]["cwd"] == PurePosixPath("/project")
    assert kwargs["container"]["oci_platform"] == OCIPlatform.i386

    identifiers = {x.identifier for x in kwargs["platform_configs"]}
    assert identifiers == {f"{x}-manylinux_i686" for x in DEFAULT_IDS}

    kwargs = build_in_container.call_args_list[2][1]
    assert "quay.io/pypa/musllinux_1_2_x86_64" in kwargs["container"]["image"]
    assert kwargs["container"]["cwd"] == PurePosixPath("/project")
    assert kwargs["container"]["oci_platform"] == OCIPlatform.AMD64

    identifiers = {x.identifier for x in kwargs["platform_configs"]}
    assert identifiers == {f"{x}-musllinux_x86_64" for x in DEFAULT_IDS}

    kwargs = build_in_container.call_args_list[3][1]
    assert "quay.io/pypa/musllinux_1_2_i686" in kwargs["container"]["image"]
    assert kwargs["container"]["cwd"] == PurePosixPath("/project")
    assert kwargs["container"]["oci_platform"] == OCIPlatform.i386

    identifiers = {x.identifier for x in kwargs["platform_configs"]}
    assert identifiers == {f"{x}-musllinux_i686" for x in DEFAULT_IDS}


@pytest.mark.usefixtures("mock_build_container")
def test_build_with_override_launches(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    pkg_dir = tmp_path / "cibw_package"
    pkg_dir.mkdir()

    cibw_toml = pkg_dir / "pyproject.toml"
    cibw_toml.write_text(
        """
[tool.cibuildwheel]
manylinux-x86_64-image = "manylinux_2_28"
musllinux-x86_64-image = "musllinux_1_2"
enable = ["pypy", "pypy-eol", "graalpy", "cpython-freethreading"]
archs = ["auto64", "auto32"]

# Before Python 3.10, use manylinux2014
[[tool.cibuildwheel.overrides]]
select = "cp3?-*"
manylinux-x86_64-image = "manylinux2014"
manylinux-i686-image = "manylinux2014"

[[tool.cibuildwheel.overrides]]
select = "cp38-manylinux_x86_64"
before-all = "true"
"""
    )

    monkeypatch.chdir(pkg_dir)
    monkeypatch.setattr(sys, "argv", ["cibuildwheel", "--platform=linux"])
    monkeypatch.delenv("CIBW_ENABLE", raising=False)

    main()

    build_in_container = typing.cast("mock.Mock", platforms.linux.build_in_container)

    assert build_in_container.call_count == 7

    kwargs = build_in_container.call_args_list[0][1]
    assert "quay.io/pypa/manylinux2014_x86_64" in kwargs["container"]["image"]
    assert kwargs["container"]["cwd"] == PurePosixPath("/project")
    assert kwargs["container"]["oci_platform"] == OCIPlatform.AMD64

    identifiers = {x.identifier for x in kwargs["platform_configs"]}
    assert identifiers == {"cp38-manylinux_x86_64"}
    assert kwargs["options"].build_options("cp38-manylinux_x86_64").before_all == "true"

    kwargs = build_in_container.call_args_list[1][1]
    assert "quay.io/pypa/manylinux2014_x86_64" in kwargs["container"]["image"]
    assert kwargs["container"]["cwd"] == PurePosixPath("/project")
    assert kwargs["container"]["oci_platform"] == OCIPlatform.AMD64

    identifiers = {x.identifier for x in kwargs["platform_configs"]}
    assert identifiers == {
        f"{x}-manylinux_x86_64"
        for x in ALL_IDS
        - {
            "cp38",
            "cp310",
            "cp311",
            "cp312",
            "cp313",
            "cp313t",
            "cp314",
            "cp314t",
            "pp38",
            "pp39",
            "pp310",
            "pp311",
            "gp311_242",
            "gp312_250",
        }
    }
    assert kwargs["options"].build_options("cp39-manylinux_x86_64").before_all == ""

    kwargs = build_in_container.call_args_list[2][1]
    assert "quay.io/pypa/manylinux_2_28_x86_64" in kwargs["container"]["image"]
    assert kwargs["container"]["cwd"] == PurePosixPath("/project")
    assert kwargs["container"]["oci_platform"] == OCIPlatform.AMD64
    identifiers = {x.identifier for x in kwargs["platform_configs"]}
    assert identifiers == {
        f"{x}-manylinux_x86_64"
        for x in [
            "cp310",
            "cp311",
            "cp312",
            "cp313",
            "cp313t",
            "cp314",
            "cp314t",
            "pp38",
            "pp39",
            "pp310",
            "pp311",
            "gp311_242",
            "gp312_250",
        ]
    }

    kwargs = build_in_container.call_args_list[3][1]
    assert "quay.io/pypa/manylinux2014_i686" in kwargs["container"]["image"]
    assert kwargs["container"]["cwd"] == PurePosixPath("/project")
    assert kwargs["container"]["oci_platform"] == OCIPlatform.i386
    identifiers = {x.identifier for x in kwargs["platform_configs"]}
    assert identifiers == {"cp38-manylinux_i686", "cp39-manylinux_i686"}

    kwargs = build_in_container.call_args_list[4][1]
    assert "quay.io/pypa/manylinux_2_28_i686" in kwargs["container"]["image"]
    assert kwargs["container"]["cwd"] == PurePosixPath("/project")
    assert kwargs["container"]["oci_platform"] == OCIPlatform.i386
    identifiers = {x.identifier for x in kwargs["platform_configs"]}
    assert identifiers == {
        f"{x}-manylinux_i686" for x in ALL_IDS - {"cp38", "cp39"} if "gp" not in x
    }

    kwargs = build_in_container.call_args_list[5][1]
    assert "quay.io/pypa/musllinux_1_2_x86_64" in kwargs["container"]["image"]
    assert kwargs["container"]["cwd"] == PurePosixPath("/project")
    assert kwargs["container"]["oci_platform"] == OCIPlatform.AMD64
    identifiers = {x.identifier for x in kwargs["platform_configs"]}
    assert identifiers == {
        f"{x}-musllinux_x86_64" for x in ALL_IDS if "pp" not in x and "gp" not in x
    }

    kwargs = build_in_container.call_args_list[6][1]
    assert "quay.io/pypa/musllinux_1_2_i686" in kwargs["container"]["image"]
    assert kwargs["container"]["cwd"] == PurePosixPath("/project")
    assert kwargs["container"]["oci_platform"] == OCIPlatform.i386

    identifiers = {x.identifier for x in kwargs["platform_configs"]}
    assert identifiers == {
        f"{x}-musllinux_i686" for x in ALL_IDS if "pp" not in x and "gp" not in x
    }



import errno
import shutil
import sys
from pathlib import Path

import pytest

import cibuildwheel.__main__ as main_module
from cibuildwheel.__main__ import main


def test_clean_cache_when_cache_exists(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch, capfd: pytest.CaptureFixture[str]
) -> None:
    fake_cache_dir = (tmp_path / "cibw_cache").resolve()
    monkeypatch.setattr(main_module, "CIBW_CACHE_PATH", fake_cache_dir)

    fake_cache_dir.mkdir(parents=True, exist_ok=True)
    assert fake_cache_dir.exists()

    cibw_sentinel = fake_cache_dir / "CACHEDIR.TAG"
    cibw_sentinel.write_text(
        "Signature: 8a477f597d28d172789f06886806bc55\n"
        "# This file is a cache directory tag created by cibuildwheel.\n"
        "# For information about cache directory tags, see:\n"
        "# https://www.brynosaurus.com/cachedir/",
        encoding="utf-8",
    )

    dummy_file = fake_cache_dir / "dummy.txt"
    dummy_file.write_text("hello")

    monkeypatch.setattr(sys, "argv", ["cibuildwheel", "--clean-cache"])

    with pytest.raises(SystemExit) as e:
        main()

    assert e.value.code == 0

    out, _ = capfd.readouterr()
    assert f"Clearing cache directory: {fake_cache_dir}" in out
    assert "Cache cleared successfully." in out
    assert not fake_cache_dir.exists()


def test_clean_cache_when_cache_does_not_exist(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch, capfd: pytest.CaptureFixture[str]
) -> None:
    fake_cache_dir = (tmp_path / "nonexistent_cache").resolve()
    monkeypatch.setattr(main_module, "CIBW_CACHE_PATH", fake_cache_dir)

    monkeypatch.setattr(sys, "argv", ["cibuildwheel", "--clean-cache"])

    with pytest.raises(SystemExit) as e:
        main()

    assert e.value.code == 0

    out, _ = capfd.readouterr()
    assert f"Cache directory does not exist: {fake_cache_dir}" in out


def test_clean_cache_with_error(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch, capfd: pytest.CaptureFixture[str]
) -> None:
    fake_cache_dir = (tmp_path / "cibw_cache").resolve()
    monkeypatch.setattr(main_module, "CIBW_CACHE_PATH", fake_cache_dir)

    fake_cache_dir.mkdir(parents=True, exist_ok=True)
    assert fake_cache_dir.exists()

    cibw_sentinel = fake_cache_dir / "CACHEDIR.TAG"
    cibw_sentinel.write_text(
        "Signature: 8a477f597d28d172789f06886806bc55\n"
        "# This file is a cache directory tag created by cibuildwheel.\n"
        "# For information about cache directory tags, see:\n"
        "# https://www.brynosaurus.com/cachedir/",
        encoding="utf-8",
    )

    monkeypatch.setattr(sys, "argv", ["cibuildwheel", "--clean-cache"])

    def fake_rmtree(path: Path) -> None:  # noqa: ARG001
        raise OSError(errno.EACCES, "Permission denied")

    monkeypatch.setattr(shutil, "rmtree", fake_rmtree)

    with pytest.raises(SystemExit) as e:
        main()

    assert e.value.code == 1

    out, err = capfd.readouterr()
    assert f"Clearing cache directory: {fake_cache_dir}" in out
    assert "Error clearing cache:" in err


def test_clean_cache_without_sentinel(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch, capfd: pytest.CaptureFixture[str]
) -> None:
    fake_cache_dir = (tmp_path / "not_a_cache").resolve()
    monkeypatch.setattr(main_module, "CIBW_CACHE_PATH", fake_cache_dir)

    fake_cache_dir.mkdir(parents=True, exist_ok=True)

    monkeypatch.setattr(sys, "argv", ["cibuildwheel", "--clean-cache"])

    with pytest.raises(SystemExit) as e:
        main()

    assert e.value.code == 1

    _, err = capfd.readouterr()
    assert "does not appear to be a cibuildwheel cache directory" in err
    assert fake_cache_dir.exists()


def test_clean_cache_with_invalid_signature(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch, capfd: pytest.CaptureFixture[str]
) -> None:
    fake_cache_dir = (tmp_path / "fake_cache").resolve()
    monkeypatch.setattr(main_module, "CIBW_CACHE_PATH", fake_cache_dir)

    fake_cache_dir.mkdir(parents=True, exist_ok=True)

    cibw_sentinel = fake_cache_dir / "CACHEDIR.TAG"
    cibw_sentinel.write_text("Invalid signature\n# This is not a real cache directory tag")

    monkeypatch.setattr(sys, "argv", ["cibuildwheel", "--clean-cache"])

    with pytest.raises(SystemExit) as e:
        main()

    assert e.value.code == 1

    _, err = capfd.readouterr()
    assert "does not contain a valid cache directory signature" in err
    assert fake_cache_dir.exists()



import inspect
import subprocess
import textwrap
from pathlib import Path

import pytest

from . import test_projects, utils

project_with_a_test = test_projects.new_c_project(
    setup_cfg_add=textwrap.dedent(
        r"""
        [options.extras_require]
        test = pytest
        """
    )
)

project_with_a_test.files["test/spam_test.py"] = r'''
import os
import platform
import sys
import struct
from unittest import TestCase

import spam


def path_contains(parent, child):
    """ returns True if `child` is inside `parent`.
    Works around path-comparison bugs caused by short-paths on Windows e.g.
    vssadm~1 instead of vssadministrator
    """
    parent = os.path.abspath(parent)
    child = os.path.abspath(child)

    while child != os.path.dirname(child):
        child = os.path.dirname(child)
        if os.stat(parent) == os.stat(child):
            # parent and child refer to the same directory on the filesystem
            return True
    return False


class TestSpam(TestCase):
    def test_filter(self) -> None:
        self.assertEqual(0, spam.filter("spam"))
        self.assertNotEqual(0, spam.filter("ham"))

    def test_virtualenv(self) -> None:
        # sys.prefix is different from sys.base_prefix when running a virtualenv
        # See https://docs.python.org/3/library/venv.html, which virtualenv seems
        # to honor in recent releases
        if sys.prefix == sys.base_prefix:
            self.fail("Not running in a virtualenv")

        self.assertTrue(path_contains(sys.prefix, sys.executable))
        self.assertTrue(path_contains(sys.prefix, spam.__file__))
        self.assertIn("VIRTUAL_ENV", os.environ)

    def test_uname(self) -> None:
        if platform.system() == "Windows":
            return
        # if we're running in 32-bit Python, check that the machine is i686.
        # See #336 for more info.
        bits = struct.calcsize("P") * 8
        if bits == 32:
            self.assertIn(platform.machine(), ["i686", "armv7l","armv8l", "wasm32"])
'''


def test(tmp_path: Path) -> None:
    project_dir = tmp_path / "project"
    project_with_a_test.generate(project_dir)

    # build and test the wheels
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_TEST_REQUIRES": "pytest",
            # the 'false ||' bit is to ensure this command runs in a shell on
            # mac/linux.
            "CIBW_TEST_COMMAND": f"false || {utils.invoke_pytest()} {{project}}/test",
            "CIBW_TEST_COMMAND_WINDOWS": "COLOR 00 || pytest {project}/test",
        },
    )

    # also check that we got the right wheels
    expected_wheels = utils.expected_wheels("spam", "0.1.0")
    assert set(actual_wheels) == set(expected_wheels)


def test_extras_require(tmp_path: Path) -> None:
    project_dir = tmp_path / "project"
    project_with_a_test.generate(project_dir)

    # build and test the wheels
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_TEST_EXTRAS": "test",
            # the 'false ||' bit is to ensure this command runs in a shell on
            # mac/linux.
            "CIBW_TEST_COMMAND": f"false || {utils.invoke_pytest()} {{project}}/test",
            "CIBW_TEST_COMMAND_WINDOWS": "COLOR 00 || pytest {project}/test",
        },
        single_python=True,
    )

    # also check that we got the right wheels
    expected_wheels = utils.expected_wheels("spam", "0.1.0", single_python=True)
    assert set(actual_wheels) == set(expected_wheels)


def test_dependency_groups(tmp_path: Path) -> None:
    group_project = project_with_a_test.copy()
    group_project.files["pyproject.toml"] = inspect.cleandoc("""
        [build-system]
        requires = ["setuptools"]
        build-backend = "setuptools.build_meta"

        [dependency-groups]
        dev = ["pytest"]
        """)

    project_dir = tmp_path / "project"
    group_project.generate(project_dir)

    # build and test the wheels
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_TEST_GROUPS": "dev",
            # the 'false ||' bit is to ensure this command runs in a shell on
            # mac/linux.
            "CIBW_TEST_COMMAND": f"false || {utils.invoke_pytest()} {{project}}/test",
            "CIBW_TEST_COMMAND_WINDOWS": "COLOR 00 || pytest {project}/test",
        },
        single_python=True,
    )

    # also check that we got the right wheels
    expected_wheels = utils.expected_wheels("spam", "0.1.0", single_python=True)
    assert set(actual_wheels) == set(expected_wheels)


project_with_a_failing_test = test_projects.new_c_project()
project_with_a_failing_test.files["test/spam_test.py"] = r"""
from unittest import TestCase

class TestSpam(TestCase):
    def test_something(self) -> None:
        self.fail('this test is supposed to fail')
"""


def test_failing_test(tmp_path: Path) -> None:
    """Ensure a failing test causes cibuildwheel to error out and exit"""
    project_dir = tmp_path / "project"
    output_dir = tmp_path / "output"
    project_with_a_failing_test.generate(project_dir)

    with pytest.raises(subprocess.CalledProcessError):
        utils.cibuildwheel_run(
            project_dir,
            output_dir=output_dir,
            add_env={
                "CIBW_TEST_REQUIRES": "pytest",
                "CIBW_TEST_COMMAND": f"{utils.invoke_pytest()} {{project}}/test",
                # CPython 3.8 when running on macOS arm64 is unusual. The build
                # always runs in x86_64, so the arm64 tests are not run. See
                # #1169 for reasons why. That means the build succeeds, which
                # we don't want. So we skip that build.
                "CIBW_SKIP": "cp38-macosx_arm64",
            },
        )

    assert len(list(output_dir.iterdir())) == 0


@pytest.mark.parametrize("test_runner", ["pytest", "unittest"])
def test_bare_pytest_invocation(
    tmp_path: Path, capfd: pytest.CaptureFixture[str], test_runner: str
) -> None:
    """
    Check that if a user runs pytest in the test cwd without setting
    test-sources, it raises a helpful error
    """
    project_dir = tmp_path / "project"
    project_with_a_test.generate(project_dir)
    output_dir = tmp_path / "output"

    with pytest.raises(subprocess.CalledProcessError):
        utils.cibuildwheel_run(
            project_dir,
            output_dir=output_dir,
            add_env={
                "CIBW_TEST_REQUIRES": "pytest" if test_runner == "pytest" else "",
                "CIBW_TEST_COMMAND": (
                    "python -m pytest" if test_runner == "pytest" else "python -m unittest"
                ),
                # Skip CPython 3.8 on macOS arm64, see comment above in
                # 'test_failing_test'
                "CIBW_SKIP": "cp38-macosx_arm64",
            },
        )

    assert len(list(output_dir.iterdir())) == 0

    captured = capfd.readouterr()

    assert (
        "Please specify a path to your tests when invoking pytest using the {project} placeholder"
        in captured.out + captured.err
    )


def test_test_sources(tmp_path: Path) -> None:
    project_dir = tmp_path / "project"
    project_with_a_test.generate(project_dir)

    # build and test the wheels in the test cwd, after copying in the test sources.
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_TEST_REQUIRES": "pytest",
            "CIBW_TEST_COMMAND_WINDOWS": "pytest",
            "CIBW_TEST_COMMAND": utils.invoke_pytest(),
            "CIBW_TEST_SOURCES": "test",
        },
    )

    # also check that we got the right wheels
    expected_wheels = utils.expected_wheels("spam", "0.1.0")
    assert set(actual_wheels) == set(expected_wheels)


def test_test_environment(tmp_path: Path) -> None:
    project_dir = tmp_path / "project"
    test_projects.new_c_project().generate(project_dir)

    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_TEST_ENVIRONMENT": "MYVAR=somevalue PYTHONSAFEPATH=1",
            "CIBW_TEST_COMMAND": "python -c \"import os; assert os.environ.get('MYVAR') == 'somevalue'; assert os.environ.get('PYTHONSAFEPATH') == '1'\"",
        },
    )
    # also check that we got the right wheels
    expected_wheels = utils.expected_wheels("spam", "0.1.0")
    assert set(actual_wheels) == set(expected_wheels)



import textwrap
from pathlib import Path

import pytest

from . import test_projects, utils

pyproject_toml = r"""
[build-system]
requires = ["setuptools", "wheel"]
build-backend = "setuptools.build_meta"
"""

limited_api_project = test_projects.new_c_project(
    setup_py_add=textwrap.dedent(
        r"""
        import sysconfig

        IS_CPYTHON = sys.implementation.name == "cpython"
        Py_GIL_DISABLED = sysconfig.get_config_var("Py_GIL_DISABLED")
        CAN_USE_ABI3 = IS_CPYTHON and not Py_GIL_DISABLED
        setup_options = {}
        extension_kwargs = {}
        if CAN_USE_ABI3 and sys.version_info[:2] >= (3, 10):
            extension_kwargs["define_macros"] = [("Py_LIMITED_API", "0x030A0000")]
            extension_kwargs["py_limited_api"] = True
            setup_options = {"bdist_wheel": {"py_limited_api": "cp310"}}
    """
    ),
    setup_py_extension_args_add="**extension_kwargs",
    setup_py_setup_args_add="options=setup_options",
)

limited_api_project.files["pyproject.toml"] = pyproject_toml


def test_abi3(tmp_path: Path) -> None:
    project_dir = tmp_path / "project"
    limited_api_project.generate(project_dir)

    # build the wheels
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            # free_threaded, GraalPy, and PyPy do not have a Py_LIMITED_API equivalent, just build one of those
            # also limit the number of builds for test performance reasons
            "CIBW_BUILD": "cp39-* cp310-* pp310-* gp312_250-* cp312-* cp314t-*",
            "CIBW_ENABLE": "all",
        },
    )

    # check that the expected wheels are produced
    if utils.get_platform() == "pyodide":
        # there's only 1 possible configuration for pyodide, cp312. It builds
        # a wheel that is tagged abi3, compatible back to 3.10
        expected_wheels = utils.expected_wheels(
            "spam",
            "0.1.0",
            python_abi_tags=["cp310-abi3"],
        )
    else:
        expected_wheels = utils.expected_wheels(
            "spam",
            "0.1.0",
            python_abi_tags=[
                "cp39-cp39",
                "cp310-abi3",  # <-- ABI3, works with 3.10 and 3.12
                "cp314-cp314t",
                "pp310-pypy310_pp73",
                "graalpy312-graalpy250_312_native",
            ],
        )

    assert set(actual_wheels) == set(expected_wheels)


ctypes_project = test_projects.TestProject()
ctypes_project.files["setup.py"] = textwrap.dedent(
    """
    from setuptools import setup, Extension

    from distutils.command.build_ext import build_ext as _build_ext
    class CTypesExtension(Extension): pass
    class build_ext(_build_ext):
        def build_extension(self, ext):
            self._ctypes = isinstance(ext, CTypesExtension)
            return super().build_extension(ext)

        def get_export_symbols(self, ext):
            if self._ctypes:
                return ext.export_symbols
            return super().get_export_symbols(ext)

        def get_ext_filename(self, ext_name):
            if self._ctypes:
                return ext_name + '.so'
            return super().get_ext_filename(ext_name)

    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel
    class bdist_wheel_abi_none(_bdist_wheel):
        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            self.root_is_pure = False

        def get_tag(self):
            python, abi, plat = _bdist_wheel.get_tag(self)
            return "py3", "none", plat

    setup(
        name="ctypesexample",
        version="1.0.0",
        package_dir = {"": "src"},
        py_modules = ["ctypesexample.summing"],
        ext_modules=[
            CTypesExtension(
                "ctypesexample.csumlib",
                ["src/ctypesexample/csumlib.c"],
            ),
        ],
        cmdclass={'build_ext': build_ext, 'bdist_wheel': bdist_wheel_abi_none},
    )
    """
)
ctypes_project.files["src/ctypesexample/csumlib.c"] = textwrap.dedent(
    """
    #ifdef _WIN32
    #define LIBRARY_API __declspec(dllexport)
    #else
    #define LIBRARY_API
    #endif

    #include <stdlib.h>


    LIBRARY_API double *add_vec3(double *a, double *b)
    {
        double *res = malloc(sizeof(double) * 3);

        for (int i = 0; i < 3; ++i)
        {
            res[i] = a[i] + b[i];
        }

        return res;
    }
    """
)
ctypes_project.files["src/ctypesexample/summing.py"] = textwrap.dedent(
    """
    import ctypes
    import pathlib

    # path of the shared library
    libfile = pathlib.Path(__file__).parent / "csumlib.so"
    csumlib = ctypes.CDLL(str(libfile))

    type_vec3 = ctypes.POINTER(ctypes.c_double * 3)

    csumlib.add_vec3.restype = type_vec3
    csumlib.add_vec3.argtypes = [type_vec3, type_vec3]
    def add(a: list, b: list) -> list:
        a_p = (ctypes.c_double * 3)(*a)
        b_p = (ctypes.c_double * 3)(*b)
        r_p = csumlib.add_vec3(a_p,b_p)

        return [l for l in r_p.contents]
    """
)

ctypes_project.files["test/add_test.py"] = textwrap.dedent(
    """
    import ctypesexample.summing

    def test():
        a = [1, 2, 3]
        b = [4, 5, 6]
        assert ctypesexample.summing.add(a, b) == [5, 7, 9]
    """
)

ctypes_project.files["pyproject.toml"] = pyproject_toml


def test_abi_none(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    project_dir = tmp_path / "project"
    ctypes_project.generate(project_dir)

    # build the wheels
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_TEST_REQUIRES": "pytest",
            "CIBW_TEST_COMMAND": f"{utils.invoke_pytest()} {{project}}/test",
            # limit the number of builds for test performance reasons
            "CIBW_BUILD": "cp38-* cp{}{}-* cp314t-* pp310-*".format(*utils.SINGLE_PYTHON_VERSION),
            "CIBW_ENABLE": "all",
        },
    )

    expected_wheels = utils.expected_wheels("ctypesexample", "1.0.0", python_abi_tags=["py3-none"])
    # check that the expected wheels are produced
    assert set(actual_wheels) == set(expected_wheels)

    captured = capfd.readouterr()

    if utils.get_platform() == "pyodide":
        # pyodide builds a different platform tag for each python version, so
        # wheels are not reused
        assert "Found previously built wheel" not in captured.out
    else:
        # check that each wheel was built once, and reused
        assert "Building wheel..." in captured.out
        assert "Found previously built wheel" in captured.out



import textwrap
from pathlib import Path

import packaging.utils
import pytest

from cibuildwheel.logger import Logger
from cibuildwheel.selector import EnableGroup

from . import test_projects, utils

basic_project = test_projects.new_c_project(
    setup_py_add=textwrap.dedent(
        """
        import os

        if os.environ.get("CIBUILDWHEEL", "0") != "1":
            raise Exception("CIBUILDWHEEL environment variable is not set to 1")
        """
    )
)


@pytest.mark.serial
def test_dummy_serial() -> None:
    """A no-op test to ensure that at least one serial test is always found.

    Without this no-op test, CI fails on CircleCI because no serial tests are
    found, and pytest errors if a test suite finds no tests.
    """


def test(
    tmp_path: Path, build_frontend_env: dict[str, str], capfd: pytest.CaptureFixture[str]
) -> None:
    project_dir = tmp_path / "project"
    basic_project.generate(project_dir)

    # build the wheels
    actual_wheels = utils.cibuildwheel_run(project_dir, add_env=build_frontend_env)

    # check that the expected wheels are produced
    expected_wheels = utils.expected_wheels("spam", "0.1.0")
    actual_wheels_normalized = {packaging.utils.parse_wheel_filename(w) for w in actual_wheels}
    expected_wheels_normalized = {packaging.utils.parse_wheel_filename(w) for w in expected_wheels}
    assert actual_wheels_normalized == expected_wheels_normalized

    enable_groups = utils.get_enable_groups()
    if EnableGroup.GraalPy not in enable_groups:
        # Verify pip warning not shown
        captured = capfd.readouterr()
        for stream in (captured.err, captured.out):
            assert "WARNING: Running pip as the 'root' user can result" not in stream
            assert "A new release of pip available" not in stream


@pytest.mark.skip(reason="to keep test output clean")
def test_sample_build(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    project_dir = tmp_path / "project"
    basic_project.generate(project_dir)

    # build the wheels, and let the output passthrough to the caller, so
    # we can see how it looks
    with capfd.disabled():
        logger = Logger()
        logger.step("test_sample_build")
        try:
            utils.cibuildwheel_run(project_dir)
        finally:
            logger.step_end()


@pytest.mark.parametrize(
    "enable_setting", ["", "cpython-prerelease", "pypy", "cpython-freethreading"]
)
def test_build_identifiers(
    tmp_path: Path, enable_setting: str, monkeypatch: pytest.MonkeyPatch
) -> None:
    project_dir = tmp_path / "project"
    basic_project.generate(project_dir)

    monkeypatch.setenv("CIBW_ENABLE", enable_setting)

    # check that the number of expected wheels matches the number of build
    # identifiers
    expected_wheels = utils.expected_wheels("spam", "0.1.0")
    build_identifiers = utils.cibuildwheel_get_build_identifiers(project_dir)
    assert len(expected_wheels) == len(build_identifiers), (
        f"{expected_wheels} vs {build_identifiers}"
    )


@pytest.mark.parametrize(
    ("add_args", "env_allow_empty"),
    [
        (["--allow-empty"], {}),
        (["--allow-empty"], {"CIBW_ALLOW_EMPTY": "0"}),
        (None, {"CIBW_ALLOW_EMPTY": "1"}),
    ],
)
def test_allow_empty(
    tmp_path: Path, add_args: list[str] | None, env_allow_empty: dict[str, str]
) -> None:
    project_dir = tmp_path / "project"
    basic_project.generate(project_dir)

    # Sanity check - --allow-empty should cause a no-op build to complete
    # without error
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={"CIBW_SKIP": "*", **env_allow_empty},
        add_args=add_args,
    )

    # check that nothing was built
    assert len(actual_wheels) == 0



import dataclasses
import json
import os
import platform as platform_module
import shutil
import subprocess
import textwrap
from collections.abc import MutableMapping, Sequence, Set
from functools import cache
from pathlib import Path
from typing import assert_never

from filelock import FileLock

from cibuildwheel import errors
from cibuildwheel.architecture import Architecture
from cibuildwheel.environment import ParsedEnvironment
from cibuildwheel.frontend import BuildFrontendName, get_build_frontend_extra_flags
from cibuildwheel.logger import log
from cibuildwheel.options import Options
from cibuildwheel.selector import BuildSelector
from cibuildwheel.util import resources
from cibuildwheel.util.cmd import call, shell
from cibuildwheel.util.file import (
    CIBW_CACHE_PATH,
    copy_test_sources,
    download,
    extract_zip,
    move_file,
)
from cibuildwheel.util.helpers import prepare_command, unwrap
from cibuildwheel.util.packaging import find_compatible_wheel, get_pip_version
from cibuildwheel.venv import constraint_flags, find_uv, virtualenv


def get_nuget_args(
    version: str, arch: str, free_threaded: bool, output_directory: Path
) -> list[str]:
    package_name = {
        "32": "pythonx86",
        "64": "python",
        "ARM64": "pythonarm64",
        # Aliases for platform.machine() return values
        "x86": "pythonx86",
        "AMD64": "python",
    }[arch]
    if free_threaded:
        package_name = f"{package_name}-freethreaded"
    return [
        package_name,
        "-Version",
        version,
        "-FallbackSource",
        "https://api.nuget.org/v3/index.json",
        "-OutputDirectory",
        str(output_directory),
    ]


@dataclasses.dataclass(frozen=True, kw_only=True)
class PythonConfiguration:
    version: str
    identifier: str
    url: str | None = None

    @property
    def arch(self) -> str:
        return {
            "win32": "32",
            "win_amd64": "64",
            "win_arm64": "ARM64",
        }[self.identifier.split("-")[-1]]


def all_python_configurations() -> list[PythonConfiguration]:
    config_dicts = resources.read_python_configs("windows")
    return [PythonConfiguration(**item) for item in config_dicts]


def get_python_configurations(
    build_selector: BuildSelector,
    architectures: Set[Architecture],
) -> list[PythonConfiguration]:
    python_configurations = all_python_configurations()

    map_arch = {"32": Architecture.x86, "64": Architecture.AMD64, "ARM64": Architecture.ARM64}

    # skip builds as required
    python_configurations = [
        c
        for c in python_configurations
        if build_selector(c.identifier) and map_arch[c.arch] in architectures
    ]

    return python_configurations


@cache
def _ensure_nuget() -> Path:
    nuget = CIBW_CACHE_PATH / "nuget.exe"
    with FileLock(str(nuget) + ".lock"):
        if not nuget.exists():
            download("https://dist.nuget.org/win-x86-commandline/latest/nuget.exe", nuget)
    return nuget


def install_cpython(configuration: PythonConfiguration, arch: str | None = None) -> Path:
    version = configuration.version
    free_threaded = "t-" in configuration.identifier
    if arch is None:
        arch = configuration.arch
    base_output_dir = CIBW_CACHE_PATH / "nuget-cpython"
    nuget_args = get_nuget_args(version, arch, free_threaded, base_output_dir)
    installation_path = base_output_dir / (nuget_args[0] + "." + version) / "tools"
    free_threaded_str = "-freethreaded" if free_threaded else ""
    with FileLock(str(base_output_dir) + f"-{version}{free_threaded_str}-{arch}.lock"):
        if not installation_path.exists():
            nuget = _ensure_nuget()
            call(nuget, "install", *nuget_args)
    return installation_path / "python.exe"


def install_pypy(tmp: Path, arch: str, url: str) -> Path:
    assert arch == "64"
    assert "win64" in url
    # Inside the PyPy zip file is a directory with the same name
    zip_filename = url.rsplit("/", 1)[-1]
    extension = ".zip"
    assert zip_filename.endswith(extension)
    installation_path = CIBW_CACHE_PATH / zip_filename[: -len(extension)]
    with FileLock(str(installation_path) + ".lock"):
        if not installation_path.exists():
            pypy_zip = tmp / zip_filename
            download(url, pypy_zip)
            # Extract to the parent directory because the zip file still contains a directory
            extract_zip(pypy_zip, installation_path.parent)
    return installation_path / "python.exe"


def install_graalpy(tmp: Path, url: str) -> Path:
    zip_filename = url.rsplit("/", 1)[-1]
    extension = ".zip"
    assert zip_filename.endswith(extension)
    installation_path = CIBW_CACHE_PATH / zip_filename[: -len(extension)]
    with FileLock(str(installation_path) + ".lock"):
        if not installation_path.exists():
            graalpy_zip = tmp / zip_filename
            download(url, graalpy_zip)
            # Extract to the parent directory because the zip file still contains a directory
            extract_zip(graalpy_zip, installation_path.parent)
    return installation_path / "bin" / "graalpy.exe"


def setup_setuptools_cross_compile(
    tmp: Path,
    python_configuration: PythonConfiguration,
    python_libs_base: Path,
    env: MutableMapping[str, str],
) -> None:
    distutils_cfg = tmp / "extra-setup.cfg"
    env["DIST_EXTRA_CONFIG"] = str(distutils_cfg)
    log.notice(f"Setting DIST_EXTRA_CONFIG={distutils_cfg} for cross-compilation")

    # Ensure our additional import libraries are made available, and explicitly
    # set the platform name
    map_plat = {"32": "win32", "64": "win-amd64", "ARM64": "win-arm64"}
    plat_name = map_plat[python_configuration.arch]

    # Set environment variable so that setuptools._distutils.get_platform()
    # identifies the target, not the host
    vscmd_arg_tgt_arch = {"32": "x86", "64": "x64", "ARM64": "arm64"}
    current_tgt_arch = vscmd_arg_tgt_arch[python_configuration.arch]
    if (env.get("VSCMD_ARG_TGT_ARCH") or current_tgt_arch) != current_tgt_arch:
        msg = f"VSCMD_ARG_TGT_ARCH must be set to {current_tgt_arch!r}, got {env['VSCMD_ARG_TGT_ARCH']!r}. If you're setting up MSVC yourself (e.g. using vcvarsall.bat or msvc-dev-cmd), make sure to target the right architecture. Alternatively, run cibuildwheel without configuring MSVC, and let the build backend handle it."
        raise errors.FatalError(msg)
    env["VSCMD_ARG_TGT_ARCH"] = current_tgt_arch

    # (This file must be default/locale encoding, so we can't pass 'encoding')
    distutils_cfg.write_text(
        textwrap.dedent(
            f"""\
            [build]
            plat_name={plat_name}
            [build_ext]
            library_dirs={python_libs_base}
            plat_name={plat_name}
            [bdist_wheel]
            plat_name={plat_name}
            """
        )
    )

    # setuptools builds require explicit override of PYD extension
    # This is because it always gets the extension from the running
    # interpreter, and has no logic to construct it. Currently, CPython's
    # extensions follow our identifiers, but if they ever diverge in the
    # future, we will need to store new data
    log.notice(
        f"Setting SETUPTOOLS_EXT_SUFFIX=.{python_configuration.identifier}.pyd for cross-compilation"
    )
    env["SETUPTOOLS_EXT_SUFFIX"] = f".{python_configuration.identifier}.pyd"

    # Cross-compilation requires fixes that only exist in setuptools's copy of
    # distutils, so ensure that it is activated
    # Since not all projects can handle the newer distutils, display a warning
    # to help them figure out what may have gone wrong if this breaks for them
    log.notice("Setting SETUPTOOLS_USE_DISTUTILS=local as it is required for cross-compilation")
    env["SETUPTOOLS_USE_DISTUTILS"] = "local"


# These cross-compile setup functions have the same signature by design
def setup_rust_cross_compile(
    tmp: Path,  # noqa: ARG001
    python_configuration: PythonConfiguration,
    python_libs_base: Path,  # noqa: ARG001
    env: MutableMapping[str, str],
) -> None:
    # Assume that MSVC will be used, because we already know that we are
    # cross-compiling. MinGW users can set CARGO_BUILD_TARGET themselves
    # and we will respect the existing value.
    cargo_target = {
        "64": "x86_64-pc-windows-msvc",
        "32": "i686-pc-windows-msvc",
        "ARM64": "aarch64-pc-windows-msvc",
    }.get(python_configuration.arch)

    # CARGO_BUILD_TARGET is the variable used by Cargo and setuptools_rust
    if env.get("CARGO_BUILD_TARGET"):
        if env["CARGO_BUILD_TARGET"] != cargo_target:
            log.notice("Not overriding CARGO_BUILD_TARGET as it has already been set")
        # No message if it was set to what we were planning to set it to
    elif cargo_target:
        log.notice(f"Setting CARGO_BUILD_TARGET={cargo_target} for cross-compilation")
        env["CARGO_BUILD_TARGET"] = cargo_target
    else:
        log.warning(
            f"Unable to configure Rust cross-compilation for architecture {python_configuration.arch}"
        )


def can_use_uv(python_configuration: PythonConfiguration) -> bool:
    conditions = (not python_configuration.identifier.startswith("pp38-"),)
    return all(conditions)


def setup_python(
    tmp: Path,
    python_configuration: PythonConfiguration,
    dependency_constraint: Path | None,
    environment: ParsedEnvironment,
    build_frontend: BuildFrontendName,
) -> tuple[Path, dict[str, str]]:
    tmp.mkdir()
    implementation_id = python_configuration.identifier.split("-")[0]
    python_libs_base = None
    log.step(f"Installing Python {implementation_id}...")
    if implementation_id.startswith("cp"):
        native_arch = platform_module.machine()
        base_python = install_cpython(python_configuration)
        if python_configuration.arch == "ARM64" != native_arch:
            # To cross-compile for ARM64, we need a native CPython to run the
            # build, and a copy of the ARM64 import libraries ('.\libs\*.lib')
            # for any extension modules.
            python_libs_base = base_python.parent / "libs"
            log.step(f"Installing native Python {native_arch} for cross-compilation...")
            base_python = install_cpython(python_configuration, arch=native_arch)
    elif implementation_id.startswith("pp"):
        assert python_configuration.url is not None
        base_python = install_pypy(tmp, python_configuration.arch, python_configuration.url)
    elif implementation_id.startswith("gp"):
        base_python = install_graalpy(tmp, python_configuration.url or "")
    else:
        msg = "Unknown Python implementation"
        raise ValueError(msg)
    assert base_python.exists()

    if build_frontend == "build[uv]" and not can_use_uv(python_configuration):
        build_frontend = "build"

    use_uv = build_frontend in {"build[uv]", "uv"}
    uv_path = find_uv()

    log.step("Setting up build environment...")
    venv_path = tmp / "venv"
    env = virtualenv(
        python_configuration.version,
        base_python,
        venv_path,
        dependency_constraint,
        use_uv=use_uv,
    )

    # set up environment variables for run_with_env
    env["PYTHON_VERSION"] = python_configuration.version
    env["PYTHON_ARCH"] = python_configuration.arch
    env.pop("UV_PYTHON", None)
    if not use_uv:
        env["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"

    # update env with results from CIBW_ENVIRONMENT
    env = environment.as_dictionary(prev_environment=env)

    # check what Python version we're on
    where_python = call("where", "python", env=env, capture_stdout=True).splitlines()[0].strip()
    print(where_python)
    if where_python != str(venv_path / "Scripts" / "python.exe"):
        msg = "python available on PATH doesn't match our installed instance. If you have modified PATH, ensure that you don't overwrite cibuildwheel's entry or insert python above it."
        raise errors.FatalError(msg)
    call("python", "--version", env=env)
    call("python", "-c", "\"import struct; print(struct.calcsize('P') * 8)\"", env=env)

    # check what pip version we're on
    if not use_uv:
        assert (venv_path / "Scripts" / "pip.exe").exists()
        where_pip = call("where", "pip", env=env, capture_stdout=True).splitlines()[0].strip()
        print(where_pip)
        if where_pip.strip() != str(venv_path / "Scripts" / "pip.exe"):
            msg = "pip available on PATH doesn't match our installed instance. If you have modified PATH, ensure that you don't overwrite cibuildwheel's entry or insert pip above it."
            raise errors.FatalError(msg)
        call("pip", "--version", env=env)

    log.step("Installing build tools...")
    match build_frontend:
        case "build":
            call(
                "pip",
                "install",
                "--upgrade",
                "build[virtualenv]",
                *constraint_flags(dependency_constraint),
                env=env,
            )
        case "build[uv]":
            assert uv_path is not None
            call(
                uv_path,
                "pip",
                "install",
                "--python",
                where_python,
                "--upgrade",
                "build[virtualenv]",
                *constraint_flags(dependency_constraint),
                env=env,
            )

    if python_libs_base:
        # Set up the environment for various backends to enable cross-compilation
        setup_setuptools_cross_compile(tmp, python_configuration, python_libs_base, env)
        setup_rust_cross_compile(tmp, python_configuration, python_libs_base, env)

    if implementation_id.startswith("gp311"):
        # GraalPy 24 fails to discover compilers, setup the relevant environment
        # variables. Adapted from
        # https://github.com/microsoft/vswhere/wiki/Start-Developer-Command-Prompt
        # Remove when GraalPy 24.x is dropped.
        vcpath = call(
            Path(os.environ["PROGRAMFILES(X86)"])
            / "Microsoft Visual Studio"
            / "Installer"
            / "vswhere.exe",
            "-products",
            "*",
            "-latest",
            "-property",
            "installationPath",
            capture_stdout=True,
        ).strip()
        log.notice(f"Discovering Visual Studio for GraalPy at {vcpath}")
        vcvars_file = tmp / "vcvars.json"
        call(
            f"{vcpath}\\Common7\\Tools\\vsdevcmd.bat",
            "-no_logo",
            "-arch=amd64",
            "-host_arch=amd64",
            "&&",
            "python",
            "-c",
            # this command needs to be one line for Windows reasons
            "import sys, json, pathlib, os; pathlib.Path(sys.argv[1]).write_text(json.dumps(dict(os.environ)))",
            vcvars_file,
            env=env,
        )
        with open(vcvars_file, encoding="utf-8") as f:
            vcvars = json.load(f)
        env.update(vcvars)

    return base_python, env


def build(options: Options, tmp_path: Path) -> None:
    python_configurations = get_python_configurations(
        options.globals.build_selector, options.globals.architectures
    )

    if not python_configurations:
        return

    uv_path = find_uv()

    try:
        before_all_options_identifier = python_configurations[0].identifier
        before_all_options = options.build_options(before_all_options_identifier)

        if before_all_options.before_all:
            log.step("Running before_all...")
            env = before_all_options.environment.as_dictionary(prev_environment=os.environ)
            before_all_prepared = prepare_command(
                before_all_options.before_all, project=".", package=options.globals.package_dir
            )
            shell(before_all_prepared, env=env)

        built_wheels: list[Path] = []

        for config in python_configurations:
            build_options = options.build_options(config.identifier)
            build_frontend = build_options.build_frontend
            use_uv = build_frontend.name in {"build[uv]", "uv"} and can_use_uv(config)
            log.build_start(config.identifier)

            identifier_tmp_dir = tmp_path / config.identifier
            identifier_tmp_dir.mkdir()
            built_wheel_dir = identifier_tmp_dir / "built_wheel"
            repaired_wheel_dir = identifier_tmp_dir / "repaired_wheel"

            constraints_path = build_options.dependency_constraints.get_for_python_version(
                version=config.version,
                tmp_dir=identifier_tmp_dir,
            )

            # install Python
            base_python, env = setup_python(
                identifier_tmp_dir / "build",
                config,
                constraints_path,
                build_options.environment,
                build_frontend.name,
            )
            pip_version = None if use_uv else get_pip_version(env)

            compatible_wheel = find_compatible_wheel(built_wheels, config.identifier)
            if compatible_wheel:
                log.step_end()
                print(
                    f"\nFound previously built wheel {compatible_wheel.name}, that's compatible with {config.identifier}. Skipping build step..."
                )
                repaired_wheel = compatible_wheel
            else:
                # run the before_build command
                if build_options.before_build:
                    log.step("Running before_build...")
                    before_build_prepared = prepare_command(
                        build_options.before_build,
                        project=".",
                        package=options.globals.package_dir,
                    )
                    shell(before_build_prepared, env=env)

                log.step("Building wheel...")
                built_wheel_dir.mkdir()

                extra_flags = get_build_frontend_extra_flags(
                    build_frontend,
                    build_options.build_verbosity,
                    build_options.config_settings,
                    py38=config.identifier[1:].startswith("p38"),
                )

                if (
                    config.identifier.startswith("gp311")
                    and build_frontend.name == "build"
                    and "--no-isolation" not in extra_flags
                    and "-n" not in extra_flags
                ):
                    # GraalPy 24 fails to discover its standard library when a venv is created
                    # from a virtualenv seeded executable. See
                    # https://github.com/oracle/graalpython/issues/491 and remove this once
                    # GraalPy 24 is dropped.
                    log.notice(
                        "Disabling build isolation to workaround GraalPy bug. If the build fails, consider using pip or build[uv] as build frontend."
                    )
                    shell("graalpy -m pip install setuptools wheel", env=env)
                    extra_flags = [*extra_flags, "-n"]

                match build_frontend.name:
                    case "pip":
                        # Path.resolve() is needed. Without it pip wheel may try to fetch package from pypi.org
                        # see https://github.com/pypa/cibuildwheel/pull/369
                        call(
                            "python",
                            "-m",
                            "pip",
                            "wheel",
                            options.globals.package_dir.resolve(),
                            f"--wheel-dir={built_wheel_dir}",
                            "--no-deps",
                            *extra_flags,
                            env=env,
                        )
                    case "build" | "build[uv]":
                        if (
                            use_uv
                            and "--no-isolation" not in extra_flags
                            and "-n" not in extra_flags
                        ):
                            extra_flags.append("--installer=uv")

                        call(
                            "python",
                            "-m",
                            "build",
                            build_options.package_dir,
                            "--wheel",
                            f"--outdir={built_wheel_dir}",
                            *extra_flags,
                            env=env,
                        )
                    case "uv":
                        assert uv_path is not None
                        call(
                            uv_path,
                            "build",
                            f"--python={base_python}",
                            build_options.package_dir,
                            "--wheel",
                            f"--out-dir={built_wheel_dir}",
                            *extra_flags,
                            env=env,
                        )
                    case _:
                        assert_never(build_frontend)

                built_wheel = next(built_wheel_dir.glob("*.whl"))

                # repair the wheel
                repaired_wheel_dir.mkdir()

                if built_wheel.name.endswith("none-any.whl"):
                    raise errors.NonPlatformWheelError()

                if build_options.repair_command:
                    log.step("Repairing wheel...")
                    repair_command_prepared = prepare_command(
                        build_options.repair_command,
                        wheel=built_wheel,
                        dest_dir=repaired_wheel_dir,
                        package=build_options.package_dir,
                        project=".",
                    )
                    shell(repair_command_prepared, env=env)
                else:
                    shutil.move(str(built_wheel), repaired_wheel_dir)

                try:
                    repaired_wheel = next(repaired_wheel_dir.glob("*.whl"))
                except StopIteration:
                    raise errors.RepairStepProducedNoWheelError() from None

                if repaired_wheel.name in {wheel.name for wheel in built_wheels}:
                    raise errors.AlreadyBuiltWheelError(repaired_wheel.name)

            test_selected = options.globals.test_selector(config.identifier)
            if test_selected and config.arch == "ARM64" != platform_module.machine():
                log.warning(
                    unwrap(
                        """
                            While arm64 wheels can be built on other platforms, they cannot
                            be tested. An arm64 runner is required. To silence this warning,
                            set `CIBW_TEST_SKIP: "*-win_arm64"`.
                            """
                    )
                )
                # skip this test
            elif test_selected and build_options.test_command:
                log.step("Testing wheel...")
                # set up a virtual environment to install and test from, to make sure
                # there are no dependencies that were pulled in at build time.
                venv_dir = identifier_tmp_dir / "venv-test"
                virtualenv_env = virtualenv(
                    config.version,
                    base_python,
                    venv_dir,
                    None,
                    use_uv=use_uv,
                    env=env,
                    pip_version=pip_version,
                )

                virtualenv_env = build_options.test_environment.as_dictionary(
                    prev_environment=virtualenv_env
                )

                # check that we are using the Python from the virtual environment
                call("where", "python", env=virtualenv_env)

                if build_options.before_test:
                    before_test_prepared = prepare_command(
                        build_options.before_test,
                        project=".",
                        package=build_options.package_dir,
                    )
                    shell(before_test_prepared, env=virtualenv_env)

                pip: Sequence[Path | str]
                if use_uv:
                    assert uv_path is not None
                    pip = [uv_path, "pip"]
                else:
                    pip = ["pip"]

                # install the wheel
                call(
                    *pip,
                    "install",
                    str(repaired_wheel) + build_options.test_extras,
                    env=virtualenv_env,
                )

                # test the wheel
                if build_options.test_requires:
                    call(*pip, "install", *build_options.test_requires, env=virtualenv_env)

                # run the tests from a temp dir, with an absolute path in the command
                # (this ensures that Python runs the tests against the installed wheel
                # and not the repo code)
                test_cwd = identifier_tmp_dir / "test_cwd"
                test_cwd.mkdir()

                if build_options.test_sources:
                    copy_test_sources(
                        build_options.test_sources,
                        Path.cwd(),
                        test_cwd,
                    )
                else:
                    # Use the test_fail.py file to raise a nice error if the user
                    # tries to run tests in the cwd
                    (test_cwd / "test_fail.py").write_text(resources.TEST_FAIL_CWD_FILE.read_text())

                test_command_prepared = prepare_command(
                    build_options.test_command,
                    project=Path.cwd(),
                    package=options.globals.package_dir.resolve(),
                    wheel=repaired_wheel,
                )
                shell(test_command_prepared, cwd=test_cwd, env=virtualenv_env)

            # we're all done here; move it to output (remove if already exists)
            output_wheel = None
            if compatible_wheel is None:
                output_wheel = build_options.output_dir.joinpath(repaired_wheel.name)
                moved_wheel = move_file(repaired_wheel, output_wheel)
                if moved_wheel != output_wheel.resolve():
                    log.warning(
                        f"{repaired_wheel} was moved to {moved_wheel} instead of {output_wheel}"
                    )
                built_wheels.append(output_wheel)

            # clean up
            # (we ignore errors because occasionally Windows fails to unlink a file and we
            # don't want to abort a build because of that)
            shutil.rmtree(identifier_tmp_dir, ignore_errors=True)

            log.build_end(output_wheel)
    except subprocess.CalledProcessError as error:
        msg = f"Command {error.cmd} failed with code {error.returncode}. {error.stdout or ''}"
        raise errors.FatalError(msg) from error



import json
from typing import Any

from cibuildwheel.util import resources


def get_schema(tool_name: str = "cibuildwheel") -> dict[str, Any]:
    "Get the stored complete schema for cibuildwheel settings."
    assert tool_name == "cibuildwheel", "Only cibuildwheel is supported."

    with resources.CIBUILDWHEEL_SCHEMA.open(encoding="utf-8") as f:
        return json.load(f)  # type: ignore[no-any-return]



import platform
import subprocess
from pathlib import Path

import pytest

from . import test_projects, utils


def test_python_exist(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    if utils.get_platform() != "linux":
        pytest.skip("the test is only relevant to the linux build")
    machine = platform.machine()
    if machine not in ["x86_64", "i686"]:
        pytest.skip(
            "this test is currently only possible on x86_64/i686 due to availability of alternative images"
        )

    project_dir = tmp_path / "project"
    basic_project = test_projects.new_c_project()
    basic_project.generate(project_dir)

    image = f"quay.io/pypa/manylinux2010_{machine}:2022-08-05-4535177"

    with pytest.raises(subprocess.CalledProcessError):
        utils.cibuildwheel_run(
            project_dir,
            add_env={
                "CIBW_MANYLINUX_X86_64_IMAGE": image,
                "CIBW_MANYLINUX_I686_IMAGE": image,
                "CIBW_BUILD": "cp3{10,11}-manylinux*",
            },
        )

    captured = capfd.readouterr()
    print("out", captured.out)
    print("err", captured.err)
    assert f" to build 'cp310-manylinux_{machine}'." not in captured.err
    message = (
        "'/opt/python/cp311-cp311/bin/python' executable doesn't exist"
        f" in image '{image}' to build 'cp311-manylinux_{machine}'."
    )
    assert message in captured.err



import functools
import tomllib
from pathlib import Path
from typing import Final

from cibuildwheel.typing import PlatformName

PATH: Final[Path] = Path(__file__).parent.parent / "resources"
INSTALL_CERTIFI_SCRIPT: Final[Path] = PATH / "install_certifi.py"
FREE_THREAD_ENABLE_313: Final[Path] = PATH / "free-threaded-enable-313.xml"
FREE_THREAD_ENABLE_314: Final[Path] = PATH / "free-threaded-enable-314.xml"
NODEJS: Final[Path] = PATH / "nodejs.toml"
DEFAULTS: Final[Path] = PATH / "defaults.toml"
PINNED_DOCKER_IMAGES: Final[Path] = PATH / "pinned_docker_images.cfg"
BUILD_PLATFORMS: Final[Path] = PATH / "build-platforms.toml"
CONSTRAINTS: Final[Path] = PATH / "constraints.txt"
VIRTUALENV: Final[Path] = PATH / "virtualenv.toml"
CIBUILDWHEEL_SCHEMA: Final[Path] = PATH / "cibuildwheel.schema.json"
PYTHON_BUILD_STANDALONE_RELEASES: Final[Path] = PATH / "python-build-standalone-releases.json"
TEST_FAIL_CWD_FILE: Final[Path] = PATH / "testing_temp_dir_file.py"


# this value is cached because it's used a lot in unit tests
@functools.cache
def read_all_configs() -> dict[str, list[dict[str, str]]]:
    with BUILD_PLATFORMS.open("rb") as f:
        loaded_file = tomllib.load(f)
    return {k: list[dict[str, str]](v["python_configurations"]) for k, v in loaded_file.items()}


def read_python_configs(config: PlatformName) -> list[dict[str, str]]:
    return read_all_configs()[config]



from pathlib import Path

import pytest

from . import test_projects, utils

basic_project = test_projects.new_c_project()


def test_wheel_tag_is_correct_when_using_macosx_deployment_target(tmp_path: Path) -> None:
    if utils.get_platform() != "macos":
        pytest.skip("This test is only relevant to MACOSX_DEPLOYMENT_TARGET")

    project_dir = tmp_path / "project"
    basic_project.generate(project_dir)

    # build the wheels
    deployment_target = "10.11"
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={"MACOSX_DEPLOYMENT_TARGET": deployment_target},
        single_python=True,
    )

    # check that the expected wheels are produced
    expected_wheels = utils.expected_wheels(
        "spam", "0.1.0", macosx_deployment_target=deployment_target, single_python=True
    )

    print("actual_wheels", actual_wheels)
    print("expected_wheels", expected_wheels)

    assert set(actual_wheels) == set(expected_wheels)



import subprocess
import textwrap
from contextlib import nullcontext as does_not_raise
from pathlib import Path

import pytest

from test import test_projects

from . import utils

basic_project = test_projects.new_c_project()
basic_project.files["repair.py"] = """
import shutil
import sys
from pathlib import Path

wheel = Path(sys.argv[1])
dest_dir = Path(sys.argv[2])
platform = wheel.stem.split("-")[-1]
if platform.startswith("pyodide"):
    # for the sake of this test, munge the pyodide platforms into one, it's
    # not valid, but it does activate the uniqueness check
    platform = "pyodide"

name = f"spam-0.1.0-py2-none-{platform}.whl"
dest = dest_dir / name
dest_dir.mkdir(parents=True, exist_ok=True)
dest.unlink(missing_ok=True)
shutil.copy(wheel, dest)
"""


def test(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    # this test checks that a generated wheel name shall be unique in a given cibuildwheel run
    project_dir = tmp_path / "project"
    basic_project.generate(project_dir)

    num_builds = len(utils.cibuildwheel_get_build_identifiers(project_dir))
    expectation = (
        pytest.raises(subprocess.CalledProcessError) if num_builds > 1 else does_not_raise()
    )

    with expectation as exc_info:
        result = utils.cibuildwheel_run(
            project_dir,
            add_env={
                "CIBW_REPAIR_WHEEL_COMMAND": "python repair.py {wheel} {dest_dir}",
            },
        )

    captured = capfd.readouterr()
    if num_builds > 1:
        assert exc_info is not None
        assert "Build failed because a wheel named" in captured.err
        assert exc_info.value.returncode == 6
    else:
        # We only produced one wheel (perhaps Pyodide)
        # check that it has the right name
        assert result[0].startswith("spam-0.1.0-py2-none-")


@pytest.mark.parametrize(
    "repair_command",
    [
        "python repair.py {wheel} {dest_dir}",
        "python {package}/repair.py {wheel} {dest_dir}",
        "python {project}/repair.py {wheel} {dest_dir}",
    ],
    ids=["no-placeholder", "package-placeholder", "project-placeholder"],
)
def test_repair_wheel_command_structure(tmp_path: Path, repair_command: str) -> None:
    project_dir = tmp_path / "project"
    project = test_projects.new_c_project()
    project.files["repair.py"] = textwrap.dedent("""
        import shutil
        import sys
        from pathlib import Path

        wheel = Path(sys.argv[1])
        dest_dir = Path(sys.argv[2])

        dest_dir.mkdir(parents=True, exist_ok=True)
        shutil.copy(wheel, dest_dir / "spamrepaired-0.0.1-py-none-any.whl")
    """)

    # Combined test for repair wheel command formats (plain, {package}, {project})
    project.generate(project_dir)

    result = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_REPAIR_WHEEL_COMMAND": repair_command,
            "CIBW_ARCHS": "native",
        },
        single_python=True,
    )

    assert result == ["spamrepaired-0.0.1-py-none-any.whl"]



#!/usr/bin/env -S uv run -q

# /// script
# dependencies = ["pyyaml"]
# ///

import argparse
import copy
import json
from typing import Any

import yaml

parser = argparse.ArgumentParser(allow_abbrev=False)
parser.add_argument("--schemastore", action="store_true", help="Generate schema_store version")
args = parser.parse_args()

starter = """
$schema: http://json-schema.org/draft-07/schema#
$id: https://github.com/pypa/cibuildwheel/blob/main/cibuildwheel/resources/cibuildwheel.schema.json
$defs:
  inherit:
    enum:
      - none
      - prepend
      - append
    default: none
    description: How to inherit the parent's value.
  enable:
    enum:
      - cpython-freethreading
      - cpython-prerelease
      - graalpy
      - pyodide-prerelease
      - pypy
      - pypy-eol
  description: A Python version or flavor to enable.
additionalProperties: false
description: cibuildwheel's settings.
type: object
properties:
  archs:
    description: Change the architectures built on your machine by default.
    type: string_array
  before-all:
    description: Execute a shell command on the build system before any wheels are built.
    type: string_array
  before-build:
    description: Execute a shell command preparing each wheel's build.
    type: string_array
  before-test:
    description: Execute a shell command before testing each wheel.
    type: string_array
  build:
    default: ['*']
    description: Choose the Python versions to build.
    type: string_array
  build-frontend:
    default: default
    description: Set the tool to use to build, either "build" (default), "build[uv]", "uv", or "pip"
    oneOf:
      - enum: [pip, build, "build[uv]", uv, default]
      - type: string
        pattern: '^pip; ?args:'
      - type: string
        pattern: '^build; ?args:'
      - type: string
        pattern: '^build\\[uv\\]; ?args:'
      - type: string
        pattern: '^uv; ?args:'
      - type: object
        additionalProperties: false
        required: [name]
        properties:
          name:
            enum: [pip, build, "build[uv]", uv]
          args:
            type: array
            items:
              type: string
  build-verbosity:
    type: integer
    minimum: -3
    maximum: 3
    default: 0
    description: Increase/decrease the output of pip wheel.
  config-settings:
    description: Specify config-settings for the build backend.
    type: string_table_array
  container-engine:
    oneOf:
      - enum: [docker, podman]
      - type: string
        pattern: '^docker; ?(create_args|disable_host_mount):'
      - type: string
        pattern: '^podman; ?(create_args|disable_host_mount):'
      - type: object
        additionalProperties: false
        required: [name]
        properties:
          name:
            enum: [docker, podman]
          create-args:
            type: array
            items:
              type: string
          disable-host-mount:
            type: boolean
  dependency-versions:
    default: pinned
    description: Specify how cibuildwheel controls the versions of the tools it uses
    oneOf:
      - enum: [pinned, latest]
      - type: string
        description: Path to a file containing dependency versions, or inline package specifications, starting with "packages:"
        not:
          enum: [pinned, latest]
      - type: object
        additionalProperties: false
        properties:
          file:
            type: string
      - type: object
        additionalProperties: false
        properties:
          packages:
            type: array
            items:
              type: string
  enable:
    description: Enable or disable certain builds.
    oneOf:
      - $ref: "#/$defs/enable"
      - type: array
        items:
          $ref: "#/$defs/enable"
  environment:
    description: Set environment variables needed during the build.
    type: string_table
  environment-pass:
    description: Set environment variables on the host to pass-through to the container
      during the build.
    type: string_array
  manylinux-aarch64-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  manylinux-armv7l-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  manylinux-i686-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  manylinux-ppc64le-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  manylinux-pypy_aarch64-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  manylinux-pypy_i686-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  manylinux-pypy_x86_64-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  manylinux-riscv64-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  manylinux-s390x-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  manylinux-x86_64-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  musllinux-aarch64-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  musllinux-armv7l-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  musllinux-i686-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  musllinux-ppc64le-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  musllinux-riscv64-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  musllinux-s390x-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  musllinux-x86_64-image:
    type: string
    description: Specify alternative manylinux / musllinux container images
  xbuild-tools:
    description: Binaries on the path that should be included in an isolated cross-build environment
    type: string_array
  pyodide-version:
    type: string
    description: Specify the version of Pyodide to use
  repair-wheel-command:
    description: Execute a shell command to repair each built wheel.
    type: string_array
  skip:
    description: Choose the Python versions to skip.
    type: string_array
  test-command:
    description: Execute a shell command to test each built wheel.
    type: string_array
  test-extras:
    description: Install your wheel for testing using `extras_require`
    type: string_array
  test-sources:
    description: Test files that are required by the test environment
    type: string_array
  test-groups:
    description: Install extra groups when testing
    type: string_array
  test-requires:
    description: Install Python dependencies before running the tests
    type: string_array
  test-skip:
    description: Skip running tests on some builds.
    type: string_array
  test-environment:
    description: Set environment variables for the test environment
    type: string_table
  test-runtime:
    description: Additional configuration for the test runner
    oneOf:
      - type: string
        pattern: '^$'
      - type: object
        additionalProperties: false
      - type: string
        pattern: 'args:'
      - type: object
        additionalProperties: false
        required: [args]
        properties:
          args:
            type: array
            items:
              type: string

"""

schema = yaml.safe_load(starter)

string_array = yaml.safe_load(
    """
- type: string
- type: array
  items:
    type: string
"""
)

string_table_array = yaml.safe_load(
    """
- type: string
- type: object
  additionalProperties: false
  patternProperties:
    .+:
      oneOf:
        - type: string
        - type: array
          items:
            type: string
"""
)

string_table = yaml.safe_load(
    """
- type: string
- type: object
  additionalProperties: false
  patternProperties:
    .+:
      type: string
"""
)

for value in schema["properties"].values():
    match value:
        case {"type": "string_array"}:
            del value["type"]
            value["oneOf"] = string_array
        case {"type": "string_table"}:
            del value["type"]
            value["oneOf"] = string_table
        case {"type": "string_table_array"}:
            del value["type"]
            value["oneOf"] = string_table_array

overrides = yaml.safe_load(
    """
type: array
description: An overrides array
items:
  type: object
  required: ["select"]
  minProperties: 2
  additionalProperties: false
  properties:
    select: {}
    inherit:
      type: object
      additionalProperties: false
      properties:
        before-all: {"$ref": "#/$defs/inherit"}
        before-build: {"$ref": "#/$defs/inherit"}
        xbuild-tools: {"$ref": "#/$defs/inherit"}
        before-test: {"$ref": "#/$defs/inherit"}
        config-settings: {"$ref": "#/$defs/inherit"}
        container-engine: {"$ref": "#/$defs/inherit"}
        environment: {"$ref": "#/$defs/inherit"}
        environment-pass: {"$ref": "#/$defs/inherit"}
        repair-wheel-command: {"$ref": "#/$defs/inherit"}
        test-command: {"$ref": "#/$defs/inherit"}
        test-extras: {"$ref": "#/$defs/inherit"}
        test-sources: {"$ref": "#/$defs/inherit"}
        test-requires: {"$ref": "#/$defs/inherit"}
        test-environment: {"$ref": "#/$defs/inherit"}
        test-runtime: {"$ref": "#/$defs/inherit"}
"""
)

for key, value in schema["properties"].items():
    value["title"] = f"CIBW_{key.replace('-', '_').upper()}"

non_global_options = {k: {"$ref": f"#/properties/{k}"} for k in schema["properties"]}
del non_global_options["build"]
del non_global_options["skip"]
del non_global_options["test-skip"]
del non_global_options["enable"]

overrides["items"]["properties"]["select"]["oneOf"] = string_array
overrides["items"]["properties"] |= non_global_options.copy()

del overrides["items"]["properties"]["archs"]

not_linux = non_global_options.copy()

del not_linux["environment-pass"]
del not_linux["container-engine"]
for key in list(not_linux):
    if "linux-" in key:
        del not_linux[key]


def as_object(d: dict[str, Any]) -> dict[str, Any]:
    return {
        "type": "object",
        "additionalProperties": False,
        "properties": copy.deepcopy(d),
    }


oses = {
    "linux": as_object(non_global_options),
    "windows": as_object(not_linux),
    "macos": as_object(not_linux),
    "pyodide": as_object(not_linux),
    "android": as_object(not_linux),
    "ios": as_object(not_linux),
}

oses["linux"]["properties"]["repair-wheel-command"] = {
    **schema["properties"]["repair-wheel-command"],
    "default": "auditwheel repair -w {dest_dir} {wheel}",
}
oses["macos"]["properties"]["repair-wheel-command"] = {
    **schema["properties"]["repair-wheel-command"],
    "default": "delocate-wheel --require-archs {delocate_archs} -w {dest_dir} -v {wheel}",
}

del oses["linux"]["properties"]["dependency-versions"]

schema["properties"]["overrides"] = overrides
schema["properties"] |= oses

if args.schemastore:
    schema["$id"] = "https://json.schemastore.org/partial-cibuildwheel.json"
    schema["description"] = (
        "cibuildwheel's settings. Generated with ./bin/generate_schema.py --schemastore from cibuildwheel."
    )

print(json.dumps(schema, indent=2))



import subprocess
from pathlib import Path

import pytest

from . import utils
from .test_projects import TestProject, new_c_project

SO_FILE_WARNING = "NOTE: Shared object (.so) files found in this project."


@pytest.mark.parametrize("project_contains_so_files", [False, True])
def test_failed_build_with_so_files(
    tmp_path: Path,
    capfd: pytest.CaptureFixture[str],
    build_frontend_env: dict[str, str],
    project_contains_so_files: bool,
) -> None:
    project = TestProject()
    project.files["setup.py"] = "raise Exception('this build will fail')\n"
    if project_contains_so_files:
        project.files["libnothing.so"] = ""

    if utils.get_platform() != "linux":
        pytest.skip("this test is only relevant to the linux build")

    project_dir = tmp_path / "project"
    project.generate(project_dir)

    with pytest.raises(subprocess.CalledProcessError):
        utils.cibuildwheel_run(project_dir, add_env=build_frontend_env)

    captured = capfd.readouterr()
    print("out", captured.out)
    print("err", captured.err)

    if project_contains_so_files:
        assert SO_FILE_WARNING in captured.err
    else:
        assert SO_FILE_WARNING not in captured.err


@pytest.mark.parametrize("project_contains_so_files", [False, True])
def test_failed_repair_with_so_files(
    tmp_path: Path,
    capfd: pytest.CaptureFixture[str],
    project_contains_so_files: bool,
) -> None:
    if utils.get_platform() != "linux":
        pytest.skip("this test is only relevant to the linux build")

    project = new_c_project()

    if project_contains_so_files:
        project.files["libnothing.so"] = ""

    project_dir = tmp_path / "project"
    project.generate(project_dir)

    with pytest.raises(subprocess.CalledProcessError):
        utils.cibuildwheel_run(project_dir, add_env={"CIBW_REPAIR_WHEEL_COMMAND": "false"})

    captured = capfd.readouterr()
    print("out", captured.out)
    print("err", captured.err)

    if project_contains_so_files:
        assert SO_FILE_WARNING in captured.err
    else:
        assert SO_FILE_WARNING not in captured.err



import os
import sys
import tomllib
from collections.abc import Mapping
from fnmatch import fnmatch
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from cibuildwheel.__main__ import main
from cibuildwheel.environment import ParsedEnvironment
from cibuildwheel.frontend import _split_config_settings, parse_config_settings
from cibuildwheel.options import BuildOptions, _get_pinned_container_images
from cibuildwheel.selector import BuildSelector, EnableGroup
from cibuildwheel.util import resources
from cibuildwheel.util.packaging import DependencyConstraints

if TYPE_CHECKING:
    from .conftest import ArgsInterceptor

# CIBW_PLATFORM is tested in main_platform_test.py


def test_old_free_threaded(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
) -> None:
    monkeypatch.setenv("CIBW_FREE_THREADED_SUPPORT", "ON")

    with pytest.raises(SystemExit):
        main()

    assert (
        "CIBW_FREE_THREADED_SUPPORT environment variable is no longer supported."
        in capsys.readouterr().err
    )


@pytest.mark.usefixtures("platform")
def test_output_dir(
    intercepted_build_args: "ArgsInterceptor", monkeypatch: pytest.MonkeyPatch
) -> None:
    OUTPUT_DIR = Path("some_output_dir")

    monkeypatch.setenv("CIBW_OUTPUT_DIR", str(OUTPUT_DIR))

    main()

    assert intercepted_build_args.args[0].globals.output_dir == OUTPUT_DIR.resolve()


@pytest.mark.usefixtures("platform")
def test_output_dir_default(intercepted_build_args: "ArgsInterceptor") -> None:
    main()

    assert intercepted_build_args.args[0].globals.output_dir == Path("wheelhouse").resolve()


@pytest.mark.usefixtures("platform")
@pytest.mark.parametrize("also_set_environment", [False, True])
def test_output_dir_argument(
    also_set_environment: bool,
    intercepted_build_args: "ArgsInterceptor",
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    OUTPUT_DIR = Path("some_output_dir")

    monkeypatch.setattr(sys, "argv", [*sys.argv, "--output-dir", str(OUTPUT_DIR)])
    if also_set_environment:
        monkeypatch.setenv("CIBW_OUTPUT_DIR", "not_this_output_dir")

    main()

    assert intercepted_build_args.args[0].globals.output_dir == OUTPUT_DIR.resolve()


@pytest.mark.usefixtures("platform", "allow_empty")
def test_build_selector(
    intercepted_build_args: "ArgsInterceptor", monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setenv("CIBW_BUILD", "cp313-*")
    monkeypatch.setenv("CIBW_SKIP", "cp39-*")

    main()

    intercepted_build_selector = intercepted_build_args.args[0].globals.build_selector
    assert isinstance(intercepted_build_selector, BuildSelector)
    assert intercepted_build_selector("cp313-something-to-build")
    assert not intercepted_build_selector("cp39-something-to-skip")
    # This unit test is just testing the options of 'main'
    # Unit tests for BuildSelector are in build_selector_test.py


@pytest.mark.usefixtures("platform", "allow_empty")
def test_invalid_build_selector(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
) -> None:
    monkeypatch.setenv("CIBW_BUILD", "invalid")

    with pytest.raises(SystemExit) as e:
        main()

    assert e.value.code == 2
    _, err = capsys.readouterr()
    assert "Invalid build selector" in err


@pytest.mark.parametrize(
    ("option_name", "option_env_var"),
    [
        ("skip", "CIBW_SKIP"),
        ("test_skip", "CIBW_TEST_SKIP"),
    ],
)
@pytest.mark.usefixtures("platform", "intercepted_build_args")
def test_invalid_skip_selector(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    option_name: str,
    option_env_var: str,
) -> None:
    monkeypatch.setenv(option_env_var, "invalid")

    main()

    _, err = capsys.readouterr()
    print(err)
    assert f"Invalid {option_name} selector" in err


@pytest.mark.parametrize(
    "selector", ["*-macosx_universal2:arm64", "*-macosx_universal2:x86_64", "*-macosx_arm64"]
)
@pytest.mark.usefixtures("platform", "intercepted_build_args")
def test_valid_test_skip_selector(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str], selector: str
) -> None:
    monkeypatch.setenv("CIBW_TEST_SKIP", selector)

    main()

    _, err = capsys.readouterr()
    print(err)
    assert "Invalid test_skip selector" not in err


@pytest.mark.parametrize("selector", ["*-macosx_universal2:invalid", "*-macosx_arm64:arm64"])
@pytest.mark.usefixtures("platform", "intercepted_build_args")
def test_invalid_test_skip_selector(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str], selector: str
) -> None:
    monkeypatch.setenv("CIBW_TEST_SKIP", selector)

    main()

    _, err = capsys.readouterr()
    print(err)
    assert "Invalid test_skip selector" in err


@pytest.mark.usefixtures("platform", "intercepted_build_args")
def test_empty_selector(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("CIBW_SKIP", "*")

    with pytest.raises(SystemExit) as e:
        main()

    assert e.value.code == 3


@pytest.mark.usefixtures("platform", "intercepted_build_args")
def test_cp313t_warning1(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
) -> None:
    monkeypatch.setenv("CIBW_ENABLE", "cpython-freethreading")

    main()

    _, err = capsys.readouterr()
    print(err)
    assert "'cpython-freethreading' enable is deprecated" in err


@pytest.mark.usefixtures("platform", "intercepted_build_args")
def test_cp313t_warning2(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str], tmp_path: Path
) -> None:
    local_path = tmp_path / "tmp_project"
    os.mkdir(local_path)  # noqa:PTH102 Path.mkdir has been monkeypatched already
    local_path.joinpath("setup.py").touch()

    monkeypatch.setattr(
        sys, "argv", ["cibuildwheel", "--only", "cp313t-manylinux_x86_64", str(local_path)]
    )
    monkeypatch.setenv("CIBW_ENABLE", "cpython-freethreading")

    main()

    _, err = capsys.readouterr()
    print(err)
    assert "'cpython-freethreading' enable is deprecated" in err


@pytest.mark.usefixtures("platform", "intercepted_build_args")
def test_cp313t_warning3(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str], tmp_path: Path
) -> None:
    local_path = tmp_path / "tmp_project"
    os.mkdir(local_path)  # noqa:PTH102 Path.mkdir has been monkeypatched already
    local_path.joinpath("setup.py").touch()

    monkeypatch.setattr(
        sys, "argv", ["cibuildwheel", "--only", "cp313t-manylinux_x86_64", str(local_path)]
    )

    main()

    _, err = capsys.readouterr()
    print(err)
    assert "'cpython-freethreading' enable is deprecated" in err


@pytest.mark.usefixtures("platform", "intercepted_build_args")
def test_cp313t_warning4(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str], tmp_path: Path
) -> None:
    local_path = tmp_path / "tmp_project"
    os.mkdir(local_path)  # noqa:PTH102 Path.mkdir has been monkeypatched already
    local_path.joinpath("setup.py").touch()

    monkeypatch.setattr(
        sys, "argv", ["cibuildwheel", "--only", "cp313t-manylinux_x86_64", str(local_path)]
    )
    monkeypatch.setenv("CIBW_ENABLE", "all")

    main()

    _, err = capsys.readouterr()
    print(err)
    assert "'cpython-freethreading' enable is deprecated" in err


@pytest.mark.parametrize(
    ("architecture", "image", "full_image"),
    [
        ("x86_64", None, "quay.io/pypa/manylinux_2_28_x86_64:*"),
        ("x86_64", "manylinux2014", "quay.io/pypa/manylinux2014_x86_64:*"),
        ("x86_64", "manylinux_2_28", "quay.io/pypa/manylinux_2_28_x86_64:*"),
        ("x86_64", "manylinux_2_34", "quay.io/pypa/manylinux_2_34_x86_64:*"),
        ("x86_64", "custom_image", "custom_image"),
        ("i686", None, "quay.io/pypa/manylinux_2_28_i686:*"),
        ("i686", "manylinux2014", "quay.io/pypa/manylinux2014_i686:*"),
        ("i686", "manylinux_2_28", "quay.io/pypa/manylinux_2_28_i686:*"),
        ("i686", "custom_image", "custom_image"),
        ("pypy_x86_64", None, "quay.io/pypa/manylinux_2_28_x86_64:*"),
        ("pypy_x86_64", "manylinux2014", "quay.io/pypa/manylinux2014_x86_64:*"),
        ("pypy_x86_64", "manylinux_2_28", "quay.io/pypa/manylinux_2_28_x86_64:*"),
        ("pypy_x86_64", "manylinux_2_34", "quay.io/pypa/manylinux_2_34_x86_64:*"),
        ("pypy_x86_64", "custom_image", "custom_image"),
    ],
)
def test_manylinux_images(
    architecture: str,
    image: str | None,
    full_image: str,
    platform: str,
    intercepted_build_args: "ArgsInterceptor",
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    if image is not None:
        monkeypatch.setenv("CIBW_MANYLINUX_" + architecture.upper() + "_IMAGE", image)

    main()

    build_options = intercepted_build_args.args[0].build_options(identifier=None)

    if platform == "linux":
        assert fnmatch(
            build_options.manylinux_images[architecture],
            full_image,
        )
    else:
        assert build_options.manylinux_images is None


def get_default_repair_command(platform: str) -> str:
    if platform == "linux":
        return "auditwheel repair -w {dest_dir} {wheel}"
    elif platform == "macos":
        return "delocate-wheel --require-archs {delocate_archs} -w {dest_dir} -v {wheel}"
    elif platform in {"windows", "pyodide"}:
        return ""
    else:
        msg = f"Unknown platform: {platform!r}"
        raise ValueError(msg)


@pytest.mark.parametrize("repair_command", [None, "repair", "repair -w {dest_dir} {wheel}"])
@pytest.mark.parametrize("platform_specific", [False, True])
def test_repair_command(
    repair_command: str | None,
    platform_specific: bool,
    platform: str,
    intercepted_build_args: "ArgsInterceptor",
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    if repair_command is not None:
        if platform_specific:
            monkeypatch.setenv("CIBW_REPAIR_WHEEL_COMMAND_" + platform.upper(), repair_command)
            monkeypatch.setenv("CIBW_REPAIR_WHEEL_COMMAND", "overwritten")
        else:
            monkeypatch.setenv("CIBW_REPAIR_WHEEL_COMMAND", repair_command)

    main()

    build_options = intercepted_build_args.args[0].build_options(identifier=None)

    expected_repair = repair_command or get_default_repair_command(platform)
    assert build_options.repair_command == expected_repair


@pytest.mark.parametrize(
    "environment",
    [{}, {"something": "value"}, {"something": "value", "something_else": "other_value"}],
)
@pytest.mark.parametrize("platform_specific", [False, True])
def test_environment(
    environment: Mapping[str, str],
    platform_specific: bool,
    platform: str,
    intercepted_build_args: "ArgsInterceptor",
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    env_string = " ".join(f"{k}={v}" for k, v in environment.items())
    if platform_specific:
        monkeypatch.setenv("CIBW_ENVIRONMENT_" + platform.upper(), env_string)
        monkeypatch.setenv("CIBW_ENVIRONMENT", "overwritten")
    else:
        monkeypatch.setenv("CIBW_ENVIRONMENT", env_string)

    main()

    build_options = intercepted_build_args.args[0].build_options(identifier=None)
    intercepted_environment = build_options.environment

    assert isinstance(intercepted_environment, ParsedEnvironment)
    assert intercepted_environment.as_dictionary(prev_environment={}) == environment


@pytest.mark.parametrize("test_requires", [None, "requirement other_requirement"])
@pytest.mark.parametrize("platform_specific", [False, True])
def test_test_requires(
    test_requires: str | None,
    platform_specific: bool,
    platform: str,
    intercepted_build_args: "ArgsInterceptor",
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    if test_requires is not None:
        if platform_specific:
            monkeypatch.setenv("CIBW_TEST_REQUIRES_" + platform.upper(), test_requires)
            monkeypatch.setenv("CIBW_TEST_REQUIRES", "overwritten")
        else:
            monkeypatch.setenv("CIBW_TEST_REQUIRES", test_requires)

    main()

    build_options = intercepted_build_args.args[0].build_options(identifier=None)

    assert build_options.test_requires == (test_requires or "").split()


@pytest.mark.parametrize("test_extras", [None, "extras"])
@pytest.mark.parametrize("platform_specific", [False, True])
def test_test_extras(
    test_extras: str | None,
    platform_specific: bool,
    platform: str,
    intercepted_build_args: "ArgsInterceptor",
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    if test_extras is not None:
        if platform_specific:
            monkeypatch.setenv("CIBW_TEST_EXTRAS_" + platform.upper(), test_extras)
            monkeypatch.setenv("CIBW_TEST_EXTRAS", "overwritten")
        else:
            monkeypatch.setenv("CIBW_TEST_EXTRAS", test_extras)

    main()

    build_options = intercepted_build_args.args[0].build_options(identifier=None)

    assert build_options.test_extras == ("[" + test_extras + "]" if test_extras else "")


@pytest.mark.parametrize("test_command", [None, "test --command"])
@pytest.mark.parametrize("platform_specific", [False, True])
def test_test_command(
    test_command: str | None,
    platform_specific: bool,
    platform: str,
    intercepted_build_args: "ArgsInterceptor",
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    if test_command is not None:
        if platform_specific:
            monkeypatch.setenv("CIBW_TEST_COMMAND_" + platform.upper(), test_command)
            monkeypatch.setenv("CIBW_TEST_COMMAND", "overwritten")
        else:
            monkeypatch.setenv("CIBW_TEST_COMMAND", test_command)

    main()

    build_options = intercepted_build_args.args[0].build_options(identifier=None)

    assert build_options.test_command == (test_command or "")


@pytest.mark.parametrize("before_build", [None, "before --build"])
@pytest.mark.parametrize("platform_specific", [False, True])
def test_before_build(
    before_build: str | None,
    platform_specific: bool,
    platform: str,
    intercepted_build_args: "ArgsInterceptor",
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    if before_build is not None:
        if platform_specific:
            monkeypatch.setenv("CIBW_BEFORE_BUILD_" + platform.upper(), before_build)
            monkeypatch.setenv("CIBW_BEFORE_BUILD", "overwritten")
        else:
            monkeypatch.setenv("CIBW_BEFORE_BUILD", before_build)

    main()

    build_options = intercepted_build_args.args[0].build_options(identifier=None)
    assert build_options.before_build == (before_build or "")


@pytest.mark.parametrize("build_verbosity", [None, 0, 2, -2, 4, -4])
@pytest.mark.parametrize("platform_specific", [False, True])
def test_build_verbosity(
    build_verbosity: int | None,
    platform_specific: bool,
    platform: str,
    intercepted_build_args: "ArgsInterceptor",
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    if build_verbosity is not None:
        if platform_specific:
            monkeypatch.setenv("CIBW_BUILD_VERBOSITY_" + platform.upper(), str(build_verbosity))
            monkeypatch.setenv("CIBW_BUILD_VERBOSITY", "overwritten")
        else:
            monkeypatch.setenv("CIBW_BUILD_VERBOSITY", str(build_verbosity))

    main()
    build_options = intercepted_build_args.args[0].build_options(identifier=None)

    expected_verbosity = max(-3, min(3, int(build_verbosity or 0)))
    assert build_options.build_verbosity == expected_verbosity


@pytest.mark.parametrize("platform_specific", [False, True])
def test_config_settings(
    platform_specific: bool,
    platform: str,
    intercepted_build_args: "ArgsInterceptor",
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    config_settings = (
        'setting=value setting=value2 triplet=1 triplet=2 triplet=3 other="something else"'
    )
    if platform_specific:
        monkeypatch.setenv("CIBW_CONFIG_SETTINGS_" + platform.upper(), config_settings)
        monkeypatch.setenv("CIBW_CONFIG_SETTINGS", "a=b")
    else:
        monkeypatch.setenv("CIBW_CONFIG_SETTINGS", config_settings)

    main()
    build_options = intercepted_build_args.args[0].build_options(identifier=None)

    assert build_options.config_settings == config_settings

    assert _split_config_settings(config_settings) == [
        "-Csetting=value",
        "-Csetting=value2",
        "-Ctriplet=1",
        "-Ctriplet=2",
        "-Ctriplet=3",
        "-Cother=something else",
    ]
    assert parse_config_settings(config_settings) == {
        "setting": ["value", "value2"],
        "triplet": ["1", "2", "3"],
        "other": "something else",
    }


@pytest.mark.parametrize(
    "selector",
    [
        "CIBW_BUILD",
        "CIBW_SKIP",
        "CIBW_TEST_SKIP",
    ],
)
@pytest.mark.parametrize(
    "pattern",
    [
        "cp27-*",
        "cp35-*",
        "?p36-*",
        "?p27*",
        "?p2*",
        "?p35*",
    ],
)
@pytest.mark.usefixtures("platform", "intercepted_build_args", "allow_empty")
def test_build_selector_deprecated_error(
    monkeypatch: pytest.MonkeyPatch, selector: str, pattern: str, capsys: pytest.CaptureFixture[str]
) -> None:
    monkeypatch.setenv(selector, pattern)
    monkeypatch.delenv("CIBW_ENABLE", raising=False)

    if selector == "CIBW_BUILD":
        with pytest.raises(SystemExit) as ex:
            main()
        assert ex.value.code == 4

    else:
        main()

    stderr = capsys.readouterr().err
    series = "2" if "6" in pattern else "1"
    msg = f"cibuildwheel 3.x no longer supports Python < 3.8. Please use the {series}.x series or update"
    assert msg in stderr


@pytest.mark.parametrize("before_all", ["", None, "test text"])
@pytest.mark.parametrize("platform_specific", [False, True])
def test_before_all(
    before_all: str | None,
    platform_specific: bool,
    platform: str,
    intercepted_build_args: "ArgsInterceptor",
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    if before_all is not None:
        if platform_specific:
            monkeypatch.setenv("CIBW_BEFORE_ALL_" + platform.upper(), before_all)
            monkeypatch.setenv("CIBW_BEFORE_ALL", "overwritten")
        else:
            monkeypatch.setenv("CIBW_BEFORE_ALL", before_all)

    main()

    build_options = intercepted_build_args.args[0].build_options(identifier=None)

    assert build_options.before_all == (before_all or "")


@pytest.mark.parametrize(
    "dependency_versions",
    [None, "pinned", "latest", "FILE", "packages: pip==21.0.0"],
)
@pytest.mark.parametrize("platform_specific", [False, True])
def test_dependency_versions(
    dependency_versions: str | None,
    platform_specific: bool,
    platform: str,
    intercepted_build_args: "ArgsInterceptor",
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    option_value = dependency_versions

    if dependency_versions == "FILE":
        constraints_file = tmp_path / "constraints.txt"
        constraints_file.write_text("foo==1.2.3\nbar==4.5.6")
        option_value = str(constraints_file)

    if option_value is not None:
        if platform_specific:
            monkeypatch.setenv("CIBW_DEPENDENCY_VERSIONS_" + platform.upper(), option_value)
            monkeypatch.setenv("CIBW_DEPENDENCY_VERSIONS", "overwritten")
        else:
            monkeypatch.setenv("CIBW_DEPENDENCY_VERSIONS", option_value)

    main()

    build_options: BuildOptions = intercepted_build_args.args[0].build_options(identifier=None)
    dependency_constraints = build_options.dependency_constraints
    if dependency_versions is None or dependency_versions == "pinned":
        assert dependency_constraints == DependencyConstraints.pinned()
    elif dependency_versions == "latest":
        assert dependency_constraints == DependencyConstraints.latest()
    elif dependency_versions == "FILE":
        assert dependency_constraints.base_file_path
        assert option_value is not None
        assert dependency_constraints.base_file_path.samefile(Path(option_value))
    elif dependency_versions.startswith("packages:"):
        assert dependency_constraints.packages == ["pip==21.0.0"]


@pytest.mark.parametrize("method", ["unset", "command_line", "env_var"])
def test_debug_traceback(
    monkeypatch: pytest.MonkeyPatch, method: str, capfd: pytest.CaptureFixture[str]
) -> None:
    if method == "command_line":
        monkeypatch.setattr(sys, "argv", [*sys.argv, "--debug-traceback"])
    elif method == "env_var":
        monkeypatch.setenv("CIBW_DEBUG_TRACEBACK", "TRUE")

    # set an option that produces a configuration error
    monkeypatch.setenv("CIBW_BUILD_FRONTEND", "invalid_value")

    with pytest.raises(SystemExit) as exit:
        main()
    assert exit.value.code == 2

    _, err = capfd.readouterr()

    if method == "unset":
        assert "Traceback (most recent call last)" not in err
    else:
        assert "Traceback (most recent call last)" in err


@pytest.mark.parametrize("method", ["unset", "command_line", "env_var"])
def test_enable(
    method: str, intercepted_build_args: "ArgsInterceptor", monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.delenv("CIBW_ENABLE", raising=False)

    if method == "command_line":
        monkeypatch.setattr(sys, "argv", [*sys.argv, "--enable", "pypy", "--enable", "graalpy"])
    elif method == "env_var":
        monkeypatch.setenv("CIBW_ENABLE", "pypy graalpy")

    main()

    enable_groups = intercepted_build_args.args[0].globals.build_selector.enable

    if method == "unset":
        assert enable_groups == frozenset()
    else:
        assert enable_groups == frozenset([EnableGroup.PyPy, EnableGroup.GraalPy])


def test_enable_all(
    intercepted_build_args: "ArgsInterceptor", monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(sys, "argv", [*sys.argv, "--enable", "all"])
    monkeypatch.delenv("CIBW_ENABLE", raising=False)

    main()

    enable_groups = intercepted_build_args.args[0].globals.build_selector.enable
    assert enable_groups == EnableGroup.all_groups()


def test_enable_arg_inherits(
    intercepted_build_args: "ArgsInterceptor", monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setenv("CIBW_ENABLE", "pypy graalpy")
    monkeypatch.setattr(sys, "argv", [*sys.argv, "--enable", "cpython-prerelease"])

    main()

    enable_groups = intercepted_build_args.args[0].globals.build_selector.enable

    assert enable_groups == frozenset(
        (EnableGroup.PyPy, EnableGroup.GraalPy, EnableGroup.CPythonPrerelease)
    )


def test_enable_arg_error_message(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
) -> None:
    monkeypatch.setattr(sys, "argv", [*sys.argv, "--enable", "invalid_group"])

    with pytest.raises(SystemExit) as ex:
        main()
    assert ex.value.code == 2

    _, err = capsys.readouterr()
    assert "Valid group names are:" in err


def test_defaults(platform: str, intercepted_build_args: "ArgsInterceptor") -> None:
    main()

    build_options: BuildOptions = intercepted_build_args.args[0].build_options(identifier=None)
    with resources.DEFAULTS.open("rb") as f:
        defaults_toml = tomllib.load(f)

    root_defaults = defaults_toml["tool"]["cibuildwheel"]
    platform_defaults = defaults_toml["tool"]["cibuildwheel"][platform]

    defaults = {}
    defaults.update(root_defaults)
    defaults.update(platform_defaults)

    # test a few options
    assert build_options.before_all == defaults["before-all"]
    repair_wheel_default = defaults["repair-wheel-command"]
    if isinstance(repair_wheel_default, list):
        repair_wheel_default = " && ".join(repair_wheel_default)
    assert build_options.repair_command == repair_wheel_default
    assert build_options.build_frontend.name == "build"
    assert build_options.build_frontend.args == ()

    if platform == "linux":
        assert build_options.manylinux_images
        pinned_images = _get_pinned_container_images()
        default_x86_64_image = pinned_images["x86_64"][defaults["manylinux-x86_64-image"]]
        assert build_options.manylinux_images["x86_64"] == default_x86_64_image



#!/usr/bin/env python3

import configparser
import dataclasses
from pathlib import Path

import requests
from packaging.version import Version

DIR = Path(__file__).parent.resolve()
RESOURCES = DIR.parent / "cibuildwheel/resources"


@dataclasses.dataclass(frozen=True)
class Image:
    manylinux_version: str
    platforms: list[str]
    image_name: str
    tag: str | None = None  # Set this to pin the image
    use_platform_suffix: bool = False


class PyPAImage(Image):
    def __init__(self, manylinux_version: str, platforms: list[str], tag: str | None = None):
        image_name = f"quay.io/pypa/{manylinux_version}"
        super().__init__(manylinux_version, platforms, image_name, tag, True)


images = [
    # manylinux2014 images
    PyPAImage(
        "manylinux2014",
        [
            "x86_64",
            "i686",
            "aarch64",
            "ppc64le",
            "s390x",
            "pypy_x86_64",
            "pypy_i686",
            "pypy_aarch64",
        ],
    ),
    # manylinux_2_28 images
    PyPAImage(
        "manylinux_2_28",
        [
            "x86_64",
            "i686",
            "aarch64",
            "ppc64le",
            "s390x",
            "pypy_x86_64",
            "pypy_i686",
            "pypy_aarch64",
        ],
    ),
    # manylinux_2_31 images
    PyPAImage("manylinux_2_31", ["armv7l"]),
    # manylinux_2_34 images
    PyPAImage(
        "manylinux_2_34",
        [
            "x86_64",
            "i686",
            "aarch64",
            "ppc64le",
            "s390x",
            "pypy_x86_64",
            "pypy_i686",
            "pypy_aarch64",
        ],
    ),
    # manylinux_2_35 images
    PyPAImage("manylinux_2_35", ["armv7l"]),
    # manylinux_2_39 images
    PyPAImage("manylinux_2_39", ["riscv64"]),
    # musllinux_1_2 images
    PyPAImage(
        "musllinux_1_2", ["x86_64", "i686", "aarch64", "ppc64le", "s390x", "armv7l", "riscv64"]
    ),
]

config = configparser.ConfigParser()

for image in images:
    # get the tag name whose digest matches 'latest'
    if image.tag is not None:
        # image has been pinned, do not update
        tag_name = image.tag
    elif image.image_name.startswith("quay.io/"):
        _, _, repository_name = image.image_name.partition("/")
        response = requests.get(
            f"https://quay.io/api/v1/repository/{repository_name}?includeTags=true"
        )
        response.raise_for_status()
        repo_info = response.json()
        tags_dict = repo_info["tags"]

        latest_tag = tags_dict.pop("latest")
        # find the tag whose manifest matches 'latest'
        tag_name = next(
            name
            for (name, info) in tags_dict.items()
            if info["manifest_digest"] == latest_tag["manifest_digest"]
        )
    elif image.image_name.startswith("ghcr.io/"):
        repository = image.image_name[8:]
        response = requests.get(
            "https://ghcr.io/token", params={"scope": f"repository:{repository}:pull"}
        )
        response.raise_for_status()
        token = response.json()["token"]
        response = requests.get(
            f"https://ghcr.io/v2/{repository}/tags/list",
            headers={"Authorization": f"Bearer {token}"},
        )
        response.raise_for_status()
        ghcr_tags = [(Version(tag), tag) for tag in response.json()["tags"] if tag != "latest"]
        ghcr_tags.sort(reverse=True)
        tag_name = ghcr_tags[0][1]
    else:
        response = requests.get(f"https://hub.docker.com/v2/repositories/{image.image_name}/tags")
        response.raise_for_status()
        tags = response.json()["results"]

        latest_tag = next(tag for tag in tags if tag["name"] == "latest")
        # i don't know what it would mean to have multiple images per tag
        assert len(latest_tag["images"]) == 1
        digest = latest_tag["images"][0]["digest"]

        pinned_tag = next(
            tag for tag in tags if tag != latest_tag and tag["images"][0]["digest"] == digest
        )
        tag_name = pinned_tag["name"]

    for platform in image.platforms:
        if not config.has_section(platform):
            config[platform] = {}
        suffix = ""
        if image.use_platform_suffix:
            suffix = f"_{platform.removeprefix('pypy_')}"
        config[platform][image.manylinux_version] = f"{image.image_name}{suffix}:{tag_name}"

with open(RESOURCES / "pinned_docker_images.cfg", "w") as f:
    config.write(f)



from pathlib import Path

import pytest

from . import test_projects, utils

basic_project = test_projects.new_c_project()


def test_podman(
    tmp_path: Path, capfd: pytest.CaptureFixture[str], request: pytest.FixtureRequest
) -> None:
    if utils.get_platform() != "linux":
        pytest.skip("the test is only relevant to the linux build")

    if not request.config.getoption("--run-podman"):
        pytest.skip("needs --run-podman option to run")

    project_dir = tmp_path / "project"
    basic_project.generate(project_dir)

    # build some musllinux and manylinux wheels (ensuring that we use two containers)
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_ARCHS": "native",
            "CIBW_BEFORE_ALL": "echo 'test log statement from before-all'",
            "CIBW_CONTAINER_ENGINE": "podman",
        },
        single_python=True,
    )

    # check that the expected wheels are produced
    expected_wheels = utils.expected_wheels("spam", "0.1.0", single_python=True, single_arch=True)
    assert set(actual_wheels) == set(expected_wheels)

    # check that stdout is bring passed-though from container correctly
    captured = capfd.readouterr()
    assert "test log statement from before-all" in captured.out


def test_create_args(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    if utils.get_platform() != "linux":
        pytest.skip("the test is only relevant to the linux build")

    project_dir = tmp_path / "project"
    basic_project.generate(project_dir)

    # build a manylinux wheel, using create_args to set an environment variable
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_SKIP": "*-musllinux_*",
            "CIBW_BEFORE_ALL": "echo TEST_CREATE_ARGS is set to $TEST_CREATE_ARGS",
            "CIBW_CONTAINER_ENGINE": "docker; create_args: --env=TEST_CREATE_ARGS=itworks",
        },
        single_python=True,
    )

    expected_wheels = utils.expected_wheels(
        "spam", "0.1.0", musllinux_versions=[], single_python=True
    )
    assert set(actual_wheels) == set(expected_wheels)

    captured = capfd.readouterr()
    assert "TEST_CREATE_ARGS is set to itworks" in captured.out



import contextlib
import os
import subprocess
import sys
import textwrap
from pathlib import Path

import pytest

from cibuildwheel.util.file import CIBW_CACHE_PATH

from . import test_projects, utils

pytestmark = pytest.mark.pyodide

CIBW_PLATFORM = os.environ.get("CIBW_PLATFORM", "pyodide")
if CIBW_PLATFORM != "pyodide":
    pytest.skip(f"{CIBW_PLATFORM=}", allow_module_level=True)

basic_project = test_projects.new_c_project()
basic_project.files["check_node.py"] = r"""
import sys
import shutil
from pathlib import Path
from pyodide.code import run_js


def check_node():
    # cibuildwheel adds a pinned node version to the PATH
    # check it's in the PATH then, check it's the one that runs Pyodide
    cibw_cache_path = Path(sys.argv[1]).resolve(strict=True)
    # find the node executable in PATH
    node = shutil.which("node")
    assert node is not None, "node is None"
    node_path = Path(node).resolve(strict=True)
    # it shall be in cibuildwheel cache
    assert cibw_cache_path in node_path.parents, f"{cibw_cache_path} not a parent of {node_path}"
    # find the path to the node executable that runs pyodide
    node_js = run_js("globalThis.process.execPath")
    assert node_js is not None, "node_js is None"
    node_js_path = Path(node_js).resolve(strict=True)
    # it shall be the one pinned by cibuildwheel
    assert node_js_path == node_path, f"{node_js_path} != {node_path}"


if __name__ == "__main__":
    check_node()
"""


@pytest.mark.parametrize("use_pyproject_toml", [True, False])
def test_pyodide_build(tmp_path: Path, use_pyproject_toml: bool) -> None:
    if sys.platform == "win32":
        pytest.skip("pyodide-build doesn't work correctly on Windows")

    if use_pyproject_toml:
        basic_project.files["pyproject.toml"] = textwrap.dedent(
            """
            [build-system]
            requires = ["setuptools"]
            build-backend = "setuptools.build_meta"
            """
        )

    project_dir = tmp_path / "project"
    basic_project.generate(project_dir)

    # check for node in 1 case only to reduce CI load
    add_env = {"CIBW_ENABLE": "pyodide-prerelease"}
    if use_pyproject_toml:
        add_env["CIBW_TEST_COMMAND"] = f"python {{project}}/check_node.py {CIBW_CACHE_PATH}"

    # build the wheels
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_args=["--platform", "pyodide"],
        add_env=add_env,
    )

    # check that the expected wheels are produced
    expected_wheels = [
        "spam-0.1.0-cp312-cp312-pyodide_2024_0_wasm32.whl",
        "spam-0.1.0-cp313-cp313-pyodide_2025_0_wasm32.whl",
    ]

    print("actual_wheels", actual_wheels)
    print("expected_wheels", expected_wheels)

    assert set(actual_wheels) == set(expected_wheels)


def test_pyodide_version_incompatible(tmp_path: Path, capfd: pytest.CaptureFixture[str]) -> None:
    if sys.platform == "win32":
        pytest.skip("pyodide-build doesn't work correctly on Windows")

    basic_project.generate(tmp_path)

    with pytest.raises(subprocess.CalledProcessError):
        utils.cibuildwheel_run(
            tmp_path,
            add_args=["--platform", "pyodide"],
            add_env={
                "CIBW_DEPENDENCY_VERSIONS": "packages: pyodide-build==0.31.2",
                "CIBW_PYODIDE_VERSION": "0.26.0a6",
            },
        )

    _, err = capfd.readouterr()

    assert "is not compatible with the pyodide-build version" in err


@pytest.mark.parametrize("expect_failure", [True, False])
def test_pyodide_build_and_test(tmp_path: Path, expect_failure: bool) -> None:
    if sys.platform == "win32":
        pytest.skip("pyodide-build doesn't work correctly on Windows")

    if expect_failure:
        basic_project.files["test/spam_test.py"] = textwrap.dedent(r"""
            def test_filter() -> None:
                assert 0 == 1
        """)
    else:
        basic_project.files["test/spam_test.py"] = textwrap.dedent(r"""
            import spam
            def test_filter() -> None:
                assert spam.filter("spam") == 0
        """)
    basic_project.generate(tmp_path)

    context = (
        pytest.raises(subprocess.CalledProcessError) if expect_failure else contextlib.nullcontext()
    )
    with context:
        # build the wheels
        actual_wheels = utils.cibuildwheel_run(
            tmp_path,
            add_args=["--platform", "pyodide"],
            add_env={
                "CIBW_TEST_REQUIRES": "pytest",
                "CIBW_TEST_COMMAND": "python -m pytest {project}",
                "CIBW_ENABLE": "pyodide-prerelease",
            },
        )
        # check that the expected wheels are produced
        expected_wheels = [
            "spam-0.1.0-cp312-cp312-pyodide_2024_0_wasm32.whl",
            "spam-0.1.0-cp313-cp313-pyodide_2025_0_wasm32.whl",
        ]
        assert set(actual_wheels) == set(expected_wheels)


def test_pyodide_repair_wheel(tmp_path: Path) -> None:
    if sys.platform == "win32":
        pytest.skip("pyodide-build doesn't work correctly on Windows")

    project_dir = tmp_path / "project"
    basic_project.generate(project_dir)

    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_args=["--platform", "pyodide"],
        add_env={
            "CIBW_REPAIR_WHEEL_COMMAND_PYODIDE": (
                "pyodide auditwheel repair --libdir /path/to/libraries --output-dir {dest_dir} {wheel}"
            ),
        },
        single_python=True,
    )

    # check that the expected wheels are produced
    expected_wheels = [
        "spam-0.1.0-cp312-cp312-pyodide_2024_0_wasm32.whl",
    ]
    assert set(actual_wheels) == set(expected_wheels)



import os
import typing
from typing import Final, Literal, Protocol

__all__ = (
    "PLATFORMS",
    "PathOrStr",
    "PlatformName",
)


PathOrStr = str | os.PathLike[str]


PlatformName = Literal["linux", "macos", "windows", "pyodide", "android", "ios"]
PLATFORMS: Final[frozenset[PlatformName]] = frozenset(typing.get_args(PlatformName))


class GenericPythonConfiguration(Protocol):
    @property
    def identifier(self) -> str: ...



#!/usr/bin/env python3

"""
Check known projects for usage of requires-python.

Usage:

    ./bin/inspect_all_known_projects.py --online=$GITHUB_TOKEN

This will cache the results to all_known_setup.yaml; you can reprint
the results without the `--online` setting.
"""

import ast
from collections.abc import Iterable, Iterator
from pathlib import Path

import click
import yaml
from github import Github, GithubException
from rich import print

from cibuildwheel.projectfiles import Analyzer

DIR = Path(__file__).parent.resolve()


def parse(contents: str) -> str | None:
    try:
        tree = ast.parse(contents)
        analyzer = Analyzer()
        analyzer.visit(tree)
        return analyzer.requires_python or ""
    except Exception:
        return None


def check_repo(name: str, contents: str) -> str:
    s = f"  {name}: "
    if name == "setup.py":
        if "python_requires" not in contents:
            s += "❌"
        res = parse(contents)
        if res is None:
            s += "⚠️ "
        elif res:
            s += "✅ " + res
        elif "python_requires" in contents:
            s += "☑️"

    elif name == "setup.cfg":
        s += "✅" if "python_requires" in contents else "❌"
    else:
        s += "✅" if "requires-python" in contents else "❌"

    return s


class MaybeRemote:
    github: Github | None
    contents: dict[str, dict[str, str | None]]

    def __init__(self, cached_file: Path | str, *, online: str | None) -> None:
        if online is not None:
            self.github = Github(online)
            self.contents = {
                "setup.py": {},
                "setup.cfg": {},
                "pyproject.toml": {},
            }
        else:
            self.github = None
            with open(cached_file) as f:
                self.contents = yaml.safe_load(f)

    def get(self, repo: str, filename: str) -> str | None:
        if self.github:
            try:
                gh_file = self.github.get_repo(repo).get_contents(filename)
            except GithubException:
                self.contents[filename][repo] = None
            else:
                assert not isinstance(gh_file, list)
                self.contents[filename][repo] = gh_file.decoded_content.decode(encoding="utf-8")

            return self.contents[filename][repo]
        elif repo in self.contents[filename]:
            return self.contents[filename][repo]
        else:
            msg = f"Trying to access {repo}:{filename} and not in cache, rebuild cache"
            raise RuntimeError(msg)

    def save(self, filename: Path | str) -> None:
        with open(filename, "w") as f:
            yaml.safe_dump(self.contents, f, default_flow_style=False)

    def on_each(self, repos: Iterable[str]) -> Iterator[tuple[str, str, str | None]]:
        for repo in repos:
            print(f"[bold]{repo}:")
            for filename in sorted(self.contents, reverse=True):
                yield repo, filename, self.get(repo, filename)


@click.command()
@click.option("--online", help="Set to $GITHUB_TOKEN")
def main(online: str | None) -> None:
    with open(DIR / "../docs/data/projects.yml") as f:
        known = yaml.safe_load(f)

    repos = [x["gh"] for x in known]

    ghinfo = MaybeRemote("all_known_setup.yaml", online=online)

    for _, filename, contents in ghinfo.on_each(repos):
        if contents is None:
            print(f"[red]  {filename}: Not found")
        else:
            print(check_repo(filename, contents))

    if online:
        ghinfo.save("all_known_setup.yaml")


if __name__ == "__main__":
    main()



import textwrap
from pathlib import Path

from . import test_projects, utils

project_with_skip_asserts = test_projects.new_c_project(
    setup_py_add=textwrap.dedent(
        rf"""
        if sys.implementation.name != "cpython":
            raise Exception("Only CPython shall be built")
        expected_version = {"({}, {})".format(*utils.SINGLE_PYTHON_VERSION)}
        if sys.version_info[0:2] != expected_version:
            raise Exception("CPython {{}}.{{}} should be skipped".format(*sys.version_info[0:2]))
        """
    )
)


def test(tmp_path: Path) -> None:
    project_dir = tmp_path / "project"
    project_with_skip_asserts.generate(project_dir)

    skip = " ".join(
        f"cp3{minor}-*" for minor in range(6, 30) if (3, minor) != utils.SINGLE_PYTHON_VERSION
    )

    # build the wheels
    actual_wheels = utils.cibuildwheel_run(
        project_dir,
        add_env={
            "CIBW_BUILD": "cp3*-*",
            "CIBW_SKIP": f"*t-* {skip}",
        },
    )

    # check that we got the right wheels. There should be a single version of CPython.
    expected_wheels = utils.expected_wheels("spam", "0.1.0", single_python=True)
    assert set(actual_wheels) == set(expected_wheels)



import platform
import textwrap
from pathlib import Path

import pytest

from . import test_projects, utils

# TODO: specify these at runtime according to manylinux_image
project_with_manylinux_symbols = test_projects.new_c_project(
    spam_c_top_level_add=textwrap.dedent(
        r"""
        #include <malloc.h>
        #include <stdlib.h>
        #include <stdint.h>
        #include <math.h>
        #include <pthread.h>

        #if !defined(__GLIBC_PREREQ)
        #error "Must run on a glibc linux environment"
        #endif

        #if !__GLIBC_PREREQ(2, 17)  /* manylinux2014 is glibc 2.17 */
        #error "Must run on a glibc >= 2.5 linux environment"
        #endif

        #if __GLIBC_PREREQ(2, 28)
        #include <threads.h>
        #endif
        """
    ),
    spam_c_function_add=textwrap.dedent(
        r"""
        #if __GLIBC_PREREQ(2, 34)
            // pthread_mutexattr_init was moved to libc.so.6 in manylinux_2_34+
            pthread_mutexattr_t attr;
            sts = pthread_mutexattr_init(&attr);
            if (sts == 0) {
                pthread_mutexattr_destroy(&attr);
            }
        #elif __GLIBC_PREREQ(2, 28)
            // thrd_equal & thrd_current are only available in manylinux_2_28+
            sts = thrd_equal(thrd_current(), thrd_current()) ? 0 : 1;
        #elif __GLIBC_PREREQ(2, 24)
            // nextupf is only available in manylinux_2_24+
            sts = (int)nextupf(0.0F);
        #elif __GLIBC_PREREQ(2, 17)  /* manylinux2014 is glibc 2.17 */
            // secure_getenv is only available in manylinux2014+
            sts = (int)(intptr_t)secure_getenv("NON_EXISTING_ENV_VARIABLE");
        #endif
        """
    ),
)


@pytest.mark.parametrize(
    "manylinux_image",
    [
        "manylinux2014",
        "manylinux_2_28",
        "manylinux_2_34",
    ],
)
@pytest.mark.usefixtures("docker_cleanup")
def test(manylinux_image: str, tmp_path: Path) -> None:
    if utils.get_platform() != "linux":
        pytest.skip("the container image test is only relevant to the linux build")

    project_dir = tmp_path / "project"
    project_with_manylinux_symbols.generate(project_dir)

    # build the wheels
    # CFLAGS environment variable is necessary to fail at build time,
    # rather than when dynamically loading the Python
    add_env = {
        "CIBW_BUILD": "*-manylinux*",
        "CIBW_ENVIRONMENT": 'CFLAGS="$CFLAGS -O0 -Werror=implicit-function-declaration"',
        "CIBW_MANYLINUX_X86_64_IMAGE": manylinux_image,
        "CIBW_MANYLINUX_I686_IMAGE": manylinux_image,
        "CIBW_MANYLINUX_PYPY_X86_64_IMAGE": manylinux_image,
        "CIBW_MANYLINUX_AARCH64_IMAGE": manylinux_image,
        "CIBW_MANYLINUX_PPC64LE_IMAGE": manylinux_image,
        "CIBW_MANYLINUX_S390X_IMAGE": manylinux_image,
        "CIBW_MANYLINUX_PYPY_AARCH64_IMAGE": manylinux_image,
        "CIBW_MANYLINUX_PYPY_I686_IMAGE": manylinux_image,
    }
    if platform.machine() == "aarch64":
        # We just have a manylinux_2_31/manylinux_2_35 image for armv7l
        add_env["CIBW_ARCHS"] = "aarch64"

    actual_wheels = utils.cibuildwheel_run(project_dir, add_env=add_env)

    platform_tag_map = {
        "manylinux2014": ["manylinux2014", "manylinux_2_17"],
    }
    expected_wheels = utils.expected_wheels(
        "spam",
        "0.1.0",
        manylinux_versions=platform_tag_map.get(manylinux_image, [manylinux_image]),
        musllinux_versions=[],
    )

    if platform.machine() == "aarch64":
        # We just have a manylinux_2_31/manylinux_2_35 image for armv7l
        expected_wheels = [w for w in expected_wheels if "armv7l" not in w]

    assert set(actual_wheels) == set(expected_wheels)



import argparse
import functools
import importlib
import subprocess
import sys
import tempfile
from pathlib import Path


def main() -> None:
    make_parser = functools.partial(argparse.ArgumentParser, allow_abbrev=False)
    if sys.version_info >= (3, 14):
        make_parser = functools.partial(make_parser, color=True, suggest_on_error=True)
    parser = make_parser(
        prog="python -m test.test_projects", description="Generate a test project to check it out"
    )
    parser.add_argument(
        "--open",
        action="store_true",
        help="Open the generated project in a file explorer",
    )
    parser.add_argument(
        "PROJECT",
        help="Python path to a project object. E.g. test.test_0_basic.basic_project",
    )
    parser.add_argument(
        "OUTPUT",
        nargs="?",
        help="Path to output dir. If no dir is passed, a tempdir will be generated.",
    )
    options = parser.parse_args()

    module, _, name = options.PROJECT.rpartition(".")

    project = getattr(importlib.import_module(module), name)

    project_dir = Path(options.OUTPUT or tempfile.mkdtemp())
    project.generate(project_dir)

    print("Project generated at", project_dir)
    print()

    if options.open:
        if sys.platform == "darwin":
            subprocess.run(["open", "--", project_dir], check=True)
        elif sys.platform == "linux2":
            subprocess.run(["xdg-open", "--", project_dir], check=True)
        elif sys.platform == "win32":
            subprocess.run(["explorer", project_dir], check=True)


if __name__ == "__main__":
    main()



import tomllib
from pathlib import Path
from textwrap import dedent

import pytest

from cibuildwheel.projectfiles import (
    get_requires_python_str,
    resolve_dependency_groups,
    setup_py_python_requires,
)


def test_read_setup_py_simple(tmp_path: Path) -> None:
    with open(tmp_path / "setup.py", "w") as f:
        f.write(
            dedent(
                """
                from setuptools import setup

                setup(
                    name = "hello",
                    other = 23,
                    example = ["item", "other"],
                    python_requires = "1.23",
                )
                """
            )
        )

    assert setup_py_python_requires(tmp_path.joinpath("setup.py").read_text()) == "1.23"
    assert get_requires_python_str(tmp_path, {}) == "1.23"


def test_read_setup_py_if_main(tmp_path: Path) -> None:
    with open(tmp_path / "setup.py", "w") as f:
        f.write(
            dedent(
                """
                from setuptools import setup

                if __name__ == "__main__":
                    setup(
                        name = "hello",
                        other = 23,
                        example = ["item", "other"],
                        python_requires = "1.23",
                    )
                """
            )
        )

    assert setup_py_python_requires(tmp_path.joinpath("setup.py").read_text()) == "1.23"
    assert get_requires_python_str(tmp_path, {}) == "1.23"


def test_read_setup_py_if_main_reversed(tmp_path: Path) -> None:
    with open(tmp_path / "setup.py", "w") as f:
        f.write(
            dedent(
                """
                from setuptools import setup

                if "__main__" == __name__:
                    setup(
                        name = "hello",
                        other = 23,
                        example = ["item", "other"],
                        python_requires = "1.23",
                    )
                """
            )
        )

    assert setup_py_python_requires(tmp_path.joinpath("setup.py").read_text()) == "1.23"
    assert get_requires_python_str(tmp_path, {}) == "1.23"


def test_read_setup_py_if_invalid(tmp_path: Path) -> None:
    with open(tmp_path / "setup.py", "w") as f:
        f.write(
            dedent(
                """
                from setuptools import setup

                if True:
                    setup(
                        name = "hello",
                        other = 23,
                        example = ["item", "other"],
                        python_requires = "1.23",
                    )
                """
            )
        )

    assert not setup_py_python_requires(tmp_path.joinpath("setup.py").read_text())
    assert not get_requires_python_str(tmp_path, {})


def test_read_setup_py_full(tmp_path: Path) -> None:
    with open(tmp_path / "setup.py", "w", encoding="utf8") as f:
        f.write(
            dedent(
                """
                import setuptools

                setuptools.randomfunc()

                setuptools.setup(
                    name = "hello",
                    description = "≥“”ü",
                    other = 23,
                    example = ["item", "other"],
                    python_requires = "1.24",
                )
                """
            )
        )

    assert (
        setup_py_python_requires(tmp_path.joinpath("setup.py").read_text(encoding="utf8")) == "1.24"
    )
    assert get_requires_python_str(tmp_path, {}) == "1.24"


def test_read_setup_py_assign(tmp_path: Path) -> None:
    with open(tmp_path / "setup.py", "w") as f:
        f.write(
            dedent(
                """
                from setuptools import setup

                REQUIRES = "3.21"

                setuptools.setup(
                    name = "hello",
                    other = 23,
                    example = ["item", "other"],
                    python_requires = REQUIRES,
                )
                """
            )
        )

    assert setup_py_python_requires(tmp_path.joinpath("setup.py").read_text()) is None
    assert get_requires_python_str(tmp_path, {}) is None


def test_read_setup_py_None(tmp_path: Path) -> None:
    with open(tmp_path / "setup.py", "w") as f:
        f.write(
            dedent(
                """
                from setuptools import setup

                REQUIRES = None

                setuptools.setup(
                    name = "hello",
                    other = 23,
                    example = ["item", "other"],
                    python_requires = None,
                )
                """
            )
        )

    assert setup_py_python_requires(tmp_path.joinpath("setup.py").read_text()) is None
    assert get_requires_python_str(tmp_path, {}) is None


def test_read_setup_py_empty(tmp_path: Path) -> None:
    with open(tmp_path / "setup.py", "w") as f:
        f.write(
            dedent(
                """
                from setuptools import setup

                REQUIRES = "3.21"

                setuptools.setup(
                    name = "hello",
                    other = 23,
                    example = ["item", "other"],
                )
                """
            )
        )

    assert setup_py_python_requires(tmp_path.joinpath("setup.py").read_text()) is None
    assert get_requires_python_str(tmp_path, {}) is None


def test_read_setup_cfg(tmp_path: Path) -> None:
    with open(tmp_path / "setup.cfg", "w") as f:
        f.write(
            dedent(
                """
                [options]
                python_requires = 1.234
                [metadata]
                something = other
                """
            )
        )

    assert get_requires_python_str(tmp_path, {}) == "1.234"


def test_read_setup_cfg_empty(tmp_path: Path) -> None:
    with open(tmp_path / "setup.cfg", "w") as f:
        f.write(
            dedent(
                """
                [options]
                other = 1.234
                [metadata]
                something = other
                """
            )
        )

    assert get_requires_python_str(tmp_path, {}) is None


def test_read_pyproject_toml(tmp_path: Path) -> None:
    with open(tmp_path / "pyproject.toml", "w") as f:
        f.write(
            dedent(
                """
                [project]
                requires-python = "1.654"

                [tool.cibuildwheel]
                something = "other"
                """
            )
        )
    with open(tmp_path / "pyproject.toml", "rb") as f:
        pyproject_toml = tomllib.load(f)

    assert get_requires_python_str(tmp_path, pyproject_toml) == "1.654"


def test_read_pyproject_toml_empty(tmp_path: Path) -> None:
    with open(tmp_path / "pyproject.toml", "w") as f:
        f.write(
            dedent(
                """
                [project]
                other = 1.234
                """
            )
        )
    with open(tmp_path / "pyproject.toml", "rb") as f:
        pyproject_toml = tomllib.load(f)

    assert get_requires_python_str(tmp_path, pyproject_toml) is None


def test_read_dep_groups() -> None:
    pyproject_toml = {"dependency-groups": {"group1": ["pkg1", "pkg2"], "group2": ["pkg3"]}}
    assert resolve_dependency_groups(pyproject_toml) == ()
    assert resolve_dependency_groups(pyproject_toml, "group1") == ("pkg1", "pkg2")
    assert resolve_dependency_groups(pyproject_toml, "group2") == ("pkg3",)
    assert resolve_dependency_groups(pyproject_toml, "group1", "group2") == ("pkg1", "pkg2", "pkg3")


def test_dep_group_no_file_error() -> None:
    with pytest.raises(FileNotFoundError, match=r"pyproject\.toml"):
        resolve_dependency_groups(None, "test")


def test_dep_group_no_section_error() -> None:
    with pytest.raises(KeyError, match=r"pyproject\.toml"):
        resolve_dependency_groups({}, "test")



#!/usr/bin/env python3

import dataclasses
import re
from pathlib import Path
from typing import Final

DIR: Final[Path] = Path(__file__).parent.parent.resolve()
OPTIONS_MD: Final[Path] = DIR / "docs" / "options.md"

SECTION_HEADER_REGEX = re.compile(r"^## (?P<name>.*?)$", re.MULTILINE)

# https://regexr.com/8f1ff
OPTION_HEADER_REGEX = re.compile(
    r"^### (?P<name>.*?){.*#(?P<id>\S+).*}\n+> ?(?P<desc>.*)$", re.MULTILINE
)


@dataclasses.dataclass(kw_only=True)
class Option:
    name: str
    id: str
    desc: str
    section: str


def get_table() -> str:
    options_md = OPTIONS_MD.read_text(encoding="utf-8")

    sections = SECTION_HEADER_REGEX.split(options_md)[1:]

    options = []

    for section_name, section_content in zip(sections[0::2], sections[1::2], strict=True):
        for match in OPTION_HEADER_REGEX.finditer(section_content):
            option = Option(
                name=match.group("name").strip(),
                id=match.group("id").strip(),
                desc=match.group("desc").strip(),
                section=section_name.strip(),
            )
            options.append(option)

    table_md = "\n<!-- This table is auto-generated from docs/options.md by bin/readme_options_table.py -->\n\n"
    table_md += "|   | Option | Description |\n"
    table_md += "|---|---|---|\n"
    last_section: str | None = None

    for option in options:
        cells: list[str] = []

        cells.append(f"**{option.section}**" if option.section != last_section else "")
        last_section = option.section

        url = f"https://cibuildwheel.pypa.io/en/stable/options/#{option.id}"
        name = option.name.replace(", ", "<br>")  # Replace commas with line breaks
        cells.append(f"[{name}]({url})")

        cells.append(option.desc)

        table_md += "| " + " | ".join(cells) + " |\n"
    table_md += "\n"

    return table_md


if __name__ == "__main__":
    print(get_table())



import json
import os
import subprocess
from collections.abc import Generator

import pytest
from filelock import FileLock

from cibuildwheel.architecture import Architecture
from cibuildwheel.ci import detect_ci_provider
from cibuildwheel.options import CommandLineArguments, Options
from cibuildwheel.selector import EnableGroup
from cibuildwheel.typing import PLATFORMS
from cibuildwheel.venv import find_uv

from . import utils
from .utils import DEFAULT_CIBW_ENABLE, EMULATED_ARCHS, get_platform


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption(
        "--run-emulation",
        action="store",
        default=None,
        help="run emulation tests",
        choices=("all", *EMULATED_ARCHS),
    )
    parser.addoption("--run-podman", action="store_true", default=False, help="run podman tests")
    parser.addoption(
        "--run-cp38-universal2",
        action="store_true",
        default=False,
        help="macOS cp38 uses the universal2 installer",
    )
    parser.addoption(
        "--enable",
        action="store",
        default=None,
        help="Set the CIBW_ENABLE environment variable for all tests.",
    )
    parser.addoption(
        "--platform",
        action="store",
        default=None,
        help="Set the CIBW_PLATFORM environment variable for all tests.",
    )


def pytest_configure(config: pytest.Config) -> None:
    flag_enable = config.getoption("--enable")
    flag_platform = config.getoption("--platform")

    if flag_enable is not None and "CIBW_ENABLE" in os.environ:
        msg = (
            "Both --enable pytest option and CIBW_ENABLE environment variable are set. "
            "Please specify only one."
        )
        raise pytest.UsageError(msg)
    if flag_platform is not None and "CIBW_PLATFORM" in os.environ:
        msg = (
            "Both --platform pytest option and CIBW_PLATFORM environment variable are set. "
            "Please specify only one."
        )
        raise pytest.UsageError(msg)

    if flag_enable is not None:
        EnableGroup.parse_option_value(flag_enable)
        os.environ["CIBW_ENABLE"] = flag_enable
    if flag_enable is None and "CIBW_ENABLE" not in os.environ:
        # Set default value for CIBW_ENABLE
        os.environ["CIBW_ENABLE"] = DEFAULT_CIBW_ENABLE

    if flag_platform is not None:
        assert flag_platform in PLATFORMS, f"Invalid platform: {flag_platform}"
        os.environ["CIBW_PLATFORM"] = flag_platform


def docker_warmup(request: pytest.FixtureRequest) -> None:
    machine = request.config.getoption("--run-emulation", default=None)
    if machine is None:
        archs = {arch.value for arch in Architecture.auto_archs("linux")}
    elif machine == "all":
        archs = set(EMULATED_ARCHS)
    else:
        archs = {machine}

    # Only include architectures where there are missing pre-installed interpreters
    archs &= {"x86_64", "i686", "aarch64"}
    if not archs:
        return

    options = Options(
        platform="linux",
        command_line_arguments=CommandLineArguments.defaults(),
        env={},
        defaults=True,
    )
    build_options = options.build_options(None)
    assert build_options.manylinux_images is not None
    assert build_options.musllinux_images is not None
    images = [build_options.manylinux_images[arch] for arch in archs] + [
        build_options.musllinux_images[arch] for arch in archs
    ]
    command = (
        "manylinux-interpreters ensure-all &&"
        "cpython3.13 -m pip download -d /tmp setuptools wheel pytest"
    )
    for image in images:
        container_id = subprocess.run(
            ["docker", "create", image, "bash", "-c", command],
            text=True,
            check=True,
            stdout=subprocess.PIPE,
        ).stdout.strip()
        try:
            subprocess.run(["docker", "start", container_id], check=True, stdout=subprocess.DEVNULL)
            exit_code = subprocess.run(
                ["docker", "wait", container_id], text=True, check=True, stdout=subprocess.PIPE
            ).stdout.strip()
            assert exit_code == "0"
            subprocess.run(
                ["docker", "commit", container_id, image], check=True, stdout=subprocess.DEVNULL
            )
        finally:
            subprocess.run(["docker", "rm", container_id], check=True, stdout=subprocess.DEVNULL)


@pytest.fixture(scope="session", autouse=True)
def docker_warmup_fixture(
    request: pytest.FixtureRequest, tmp_path_factory: pytest.TempPathFactory, worker_id: str
) -> None:
    # if we're in CI testing linux, let's warm-up docker images
    if detect_ci_provider() is None or get_platform() != "linux":
        return None
    if request.config.getoption("--run-emulation", default=None) is not None:
        # emulation tests only run one test in CI, caching the image only slows down the test
        return None

    if worker_id == "master":
        # not executing with multiple workers
        # it might be unsafe to write to tmp_path_factory.getbasetemp().parent
        return docker_warmup(request)

    # get the temp directory shared by all workers
    root_tmp_dir = tmp_path_factory.getbasetemp().parent

    fn = root_tmp_dir / "warmup.done"
    with FileLock(str(fn) + ".lock"):
        if not fn.is_file():
            docker_warmup(request)
            fn.write_text("done")
    return None


@pytest.fixture(params=["pip", "build"])
def build_frontend_env_nouv(request: pytest.FixtureRequest) -> dict[str, str]:
    frontend = request.param
    marks = {m.name for m in request.node.iter_markers()}

    platform = "pyodide" if "pyodide" in marks else get_platform()
    if platform == "pyodide" and frontend == "pip":
        pytest.skip("Can't use pip as build frontend for pyodide platform")

    return {"CIBW_BUILD_FRONTEND": frontend}


@pytest.fixture(params=["pip", "build", "build[uv]", "uv"])
def build_frontend_env(request: pytest.FixtureRequest) -> Generator[dict[str, str], None, None]:
    frontend = request.param
    marks = {m.name for m in request.node.iter_markers()}
    if "android" in marks:
        platform = "android"
    elif "ios" in marks:
        platform = "ios"
    elif "pyodide" in marks:
        platform = "pyodide"
    else:
        platform = get_platform()

    if platform in {"pyodide", "ios", "android"} and frontend == "pip":
        pytest.skip(f"Can't use pip as build frontend for {platform}")
    if platform == "pyodide" and frontend in {"build[uv]", "uv"}:
        pytest.skip("Can't use uv with pyodide yet")
    uv_path = find_uv()
    if uv_path is None and frontend in {"build[uv]", "uv"}:
        pytest.skip("Can't find uv, so skipping uv tests")
    if uv_path is not None and frontend == "build" and platform not in {"android", "ios"}:
        pytest.skip("No need to check build when uv is present")

    # temporary workaround: uv doesn't work with graalpy yet
    uses_uv = "uv" in frontend
    env: dict[str, str] = {"CIBW_BUILD_FRONTEND": frontend}
    if uses_uv:
        utils.include_graalpy_in_expected_wheels = False
        env["CIBW_SKIP"] = "gp*"  # skip graalpy when using uv, until uv supports it
    try:
        yield env
    finally:
        if uses_uv:
            utils.include_graalpy_in_expected_wheels = True


@pytest.fixture
def docker_cleanup() -> Generator[None, None, None]:
    def get_images() -> set[str]:
        if detect_ci_provider() is None or get_platform() != "linux":
            return set()
        images = subprocess.run(
            ["docker", "image", "ls", "--format", "{{json .ID}}"],
            text=True,
            check=True,
            stdout=subprocess.PIPE,
        ).stdout
        return {json.loads(image.strip()) for image in images.splitlines() if image.strip()}

    images_before = get_images()
    try:
        yield
    finally:
        images_after = get_images()
        for image in images_after - images_before:
            subprocess.run(["docker", "rmi", image], check=False)
